#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import time
import cv2
import threading
import numpy as np

from sciveo.tools.logger import *
from sciveo.tools.timers import FPSCounter


class BaseInteractiveVideoPlayer:
  def __init__(self, path_video, tag):
    self.path_video = path_video
    self.file_name = os.path.basename(self.path_video)
    self.file_name_base = filename_without_ext = os.path.splitext(self.file_name)[0]
    self.path_base = os.path.dirname(self.path_video)
    self.tag = tag
    self.frame_tag = f"Video {tag}"

    self.cap = None
    self.frame = None
    self.frame_id = 0
    self.result = None
    self.progress_lock = threading.Lock()
    self.frame_lock = threading.Lock()
    self.ui_video_progress_tag = "video time progress"

    self.fps_ui = FPSCounter(tag="ui", period=1, print_period=600)
    self.fps_read = FPSCounter(tag="read", period=1, print_period=600)

    self.stop_threads = False
    self.threads = [
      threading.Thread(target=self.read, daemon=True)
    ]

  def callback_video_progress(self, x):
    # debug("callback_video_progress", x)
    with self.frame_lock:
      # self.frame_id = cv2.getTrackbarPos(self.ui_video_progress_tag, self.frame_tag)
      self.frame_id = x

  def ui_controls(self, min_val, max_val):
    cv2.createTrackbar(self.ui_video_progress_tag, self.frame_tag, min_val, max_val, self.callback_video_progress)

  def process_frame(self, frame):
    return {}

  def read(self):
    while not self.stop_threads:
      self.fps_read.update()
      with self.progress_lock:
        current_frame_id = self.frame_id
      self.cap.set(cv2.CAP_PROP_POS_FRAMES, int(current_frame_id + 1))
      ret, current_frame = self.cap.read()
      if not ret:
        break

      current_result = self.process_frame(current_frame)

      with self.frame_lock:
        self.frame = current_frame.copy()
        self.result = current_result

      with self.progress_lock:
        self.frame_id += 1
        if self.frame_id >= self.video_frame_count:
          self.frame_id = 0
        cv2.setTrackbarPos(self.ui_video_progress_tag, self.frame_tag, self.frame_id)

  def draw(self, frame, result):
    cv2.putText(frame, f"R[{result}]", (30, 80), cv2.FONT_HERSHEY_SIMPLEX, 1.3, (0, 255, 255), 2)

  def run(self, max_display_width=1024):
    cv2.namedWindow(self.frame_tag, cv2.WINDOW_NORMAL)

    self.cap = cv2.VideoCapture(self.path_video)

    self.video_fps = self.cap.get(cv2.CAP_PROP_FPS)
    self.video_frame_count = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
    self.video_duration_sec = self.video_frame_count / self.video_fps
    debug("run", self.path_video, self.tag, "fps", self.video_fps, "count", self.video_frame_count, "duration", self.video_duration_sec)

    self.ui_controls(0, self.video_frame_count)

    for thread in self.threads:
      thread.start()

    while self.cap.isOpened():
      self.fps_ui.update()
      current_frame = None
      original_current_frame = None
      current_result = None
      with self.frame_lock:
        if self.frame is not None:
          current_frame = self.frame.copy()
          original_current_frame = self.frame.copy()
          current_result = self.result.copy()

      if current_frame is not None:
        cv2.putText(current_frame, f"FPS read {round(self.fps_read.value, 1)} ui {round(self.fps_ui.value, 1)} T {round(self.frame_id / self.video_fps, 3)}s", (30, 40), cv2.FONT_HERSHEY_SIMPLEX, 1.3, (0, 255, 0), 4)
        self.draw(current_frame, current_result)
        height, width = current_frame.shape[:2]
        if width > max_display_width:
          display_width = max_display_width
          display_height = int(height * (max_display_width / width))
          cv2.resizeWindow(self.frame_tag, display_width, display_height)
        cv2.imshow(self.frame_tag, current_frame)

      key = cv2.waitKey(1) & 0xFF
      if key == ord("q"):
        self.stop_threads = True
        for thread in self.threads:
          thread.join()
        break
      elif key == ord("p"):
        if original_current_frame is not None:
          frame_path = os.path.join(self.path_base, f"{self.file_name_base}-{self.frame_id}.png")
          cv2.imwrite(frame_path, original_current_frame)
          debug(f"saved frame {self.frame_id} to {frame_path}")

    self.cap.release()
    cv2.destroyAllWindows()


if __name__ == "__main__":
  video_player = BaseInteractiveVideoPlayer(os.environ["PATH_VIDEO_FILE"], "base video player")
  video_player.run()


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import shutil
import yaml
import json
import time
import math
import numpy as np

from sciveo.tools.logger import *
from sciveo.ml.evaluation.object_detection import *


def xcycwh_to_xyxy(bbox):
  cx, cy, w, h = bbox
  x1 = cx - w / 2
  y1 = cy - h / 2
  x2 = cx + w / 2
  y2 = cy + h / 2
  return [x1, y1, x2, y2]

def xyxy_to_xcycwh(bbox):
  x1, y1, x2, y2 = bbox
  cx = (x1 + x2) / 2
  cy = (y1 + y2) / 2
  w = x2 - x1
  h = y2 - y1
  return [cx, cy, w, h]

def bbox_from_yolo(yolo_bbox):
  yolo_bbox = [float(a) for a in yolo_bbox]
  n, cx, cy, w, h = yolo_bbox
  return xcycwh_to_xyxy([cx, cy, w, h])

def bbox_to_yolo(bbox):
  n, x1, y1, x2, y2 = bbox
  cx, cy, w, h = xyxy_to_xcycwh([x1, y1, x2, y2])
  return [n, cx, cy, w, h]


class YOLODataset:
  def __init__(self):
    self.src_path = None
    self.dst_path = None

    self.classes = {}
    self.splits = []
    self.labels = {}

  def load(self, base_path):
    self.src_path = base_path
    yaml_path = os.path.join(base_path, "data.yaml")
    if not os.path.exists(yaml_path):
      raise FileNotFoundError(f"data.yaml not found in {base_path}")

    with open(yaml_path, "r") as f:
      self.data_yaml = yaml.safe_load(f)

    self.classes = self.data_yaml.get("names", {})
    self.splits = [key for key in self.data_yaml.keys() if key not in ["names", "nc", "path"]]
    self.labels = {split: self._load_labels_for_split(base_path, split) for split in self.splits}

  def _load_labels_for_split(self, base_path, split: str):
    images_dir = self.data_yaml.get(split)
    labels_dir = images_dir.replace("images", "labels")
    if not labels_dir:
      return {}

    images_path = os.path.join(base_path, images_dir)
    labels_path = os.path.join(base_path, labels_dir)
    if not os.path.exists(labels_path):
      return {}

    annotations = {}
    for label_file in os.listdir(labels_path):
      if label_file.endswith(".txt"):
        image_id = label_file.replace(".txt", ".jpg")
        label_path = os.path.join(labels_path, label_file)

        with open(label_path, "r") as f:
          annotations[image_id] = [line.strip().split() for line in f.readlines()]

    return annotations

  def save(self, base_path, classes=[], copy_images=True):
    self.dst_path = base_path
    os.makedirs(self.dst_path, exist_ok=True)

    yaml_path = os.path.join(base_path, "data.yaml")

    if not classes:
      selected_classes = self.classes
    else:
      selected_classes = {idx: name for idx, name in self.classes.items() if name in classes}

    class_mapping = {name: new_idx for new_idx, (idx, name) in enumerate(selected_classes.items())}

    # data_yaml = {
    #   "path": "./",
    #   "names": {idx: name for idx, name in selected_classes.items()}
    # }
    # for split in self.splits:
    #   data_yaml[split] = f"images/{split}"
    # with open(yaml_path, "w") as f:
    #   yaml.dump(data_yaml, f, default_flow_style=False)

    with open(yaml_path, "w") as fp:
      fp.write("  path: ./ # dataset root dir\n")
      for split in self.splits:
        fp.write(f"  test: images/{split}\n")
      fp.write("  names:\n")
      for class_id, class_name in selected_classes.items():
        fp.write(f"    {class_id}: {class_name}\n")

    for split, images in self.labels.items():
      labels_path = os.path.join(base_path, "labels", split)
      images_path = os.path.join(base_path, "images", split)
      os.makedirs(labels_path, exist_ok=True)
      os.makedirs(images_path, exist_ok=True)

      for image_name, annotations in images.items():
        label_file_path = os.path.join(labels_path, image_name.replace(".jpg", ".txt"))

        with open(label_file_path, "w") as f:
          for annotation in annotations:
            class_id, *bbox = map(float, annotation)
            class_id = int(class_id)

            class_name = self.classes.get(class_id)
            if class_name in class_mapping:
              new_class_id = class_mapping[class_name]
              f.write(f"{new_class_id} {' '.join(map(str, bbox))}\n")

        if copy_images and self.src_path:
          src_image_path = os.path.join(self.src_path, "images", split, image_name)
          dst_image_path = os.path.join(images_path, image_name)
          if not os.path.isfile(dst_image_path):
            shutil.copy(src_image_path, dst_image_path)


"""
Object Detection Dataset

labels in splits (for example train/val/test)
every dataset split has image name as key which points to dict of classes as keys

labels[split][image_name][class_name] is a list of bounding boxes of same class
bboxes are of type nxyxy which means normalized [x1, y1, x2, y2] coordinates (upper left and down right point of the rectangle)

"""
class ObjectDetectionDataset:
  def __init__(self):
    self.classes = []
    self.splits = []
    self.labels = {}

    self.src_path = None
    self.dst_path = None

  def save(self, base_path, copy_images=True, src_path=None):
    if src_path is not None:
      self.src_path = src_path

    self.dst_path = base_path
    for split in self.splits:
      images_path = os.path.join(self.dst_path, split, "images")
      os.makedirs(images_path, exist_ok=True)

      labels_path = os.path.join(self.dst_path, split, "labels.json")
      with open(labels_path, 'w') as fp:
        json.dump(self.labels[split], fp, indent=2)

      if copy_images and self.src_path:
        src_images_path = os.path.join(self.src_path, "images", split)
        for image_path, subdirs, image_files in os.walk(src_images_path):
          for image_name in image_files:
            src_path = os.path.join(image_path, image_name)
            dst_path = os.path.join(images_path, image_name)
            if not os.path.isfile(dst_path):
              shutil.copy(src_path, dst_path)

  def from_yolo(self, base_path):
    self.src_path = base_path

    self.raw = YOLODataset()
    self.raw.load(base_path)

    self.classes = []
    for i in range(len(self.raw.classes)):
      self.classes.append(self.raw.classes[i])
    self.splits = self.raw.splits
    self.image_labels = {}
    self.labels = {}
    for split, split_labels in self.raw.labels.items():
      self.image_labels.setdefault(split, {})
      self.labels.setdefault(split, {})
      for image_id, image_labels in split_labels.items():
        self.image_labels[split].setdefault(image_id, {})
        for label in image_labels:
          class_id = int(label[0])
          class_name = self.raw.classes[class_id]
          bbox = bbox_from_yolo(label)

          self.image_labels[split][image_id].setdefault(class_name, [])
          self.image_labels[split][image_id][class_name].append(bbox)

          self.labels[split].setdefault(class_name, {})
          self.labels[split][class_name].setdefault(image_id, [])
          self.labels[split][class_name][image_id].append(bbox)

  def stats(self):
    result = {"datasets": {}, "count": {}, "distribution": {}}
    for split in self.splits:
      result["datasets"].setdefault(split, {})
      result["count"].setdefault(split, 0)
      for class_name in self.classes:
        result["datasets"][split].setdefault(class_name, 0)
        for image_name, image_labels in self.labels[split][class_name].items():
          result["count"][split] += len(image_labels)
          result["datasets"][split][class_name] += len(image_labels)

    for split in self.splits:
      result["distribution"].setdefault(split, [])
      for class_name in self.classes:
        result["distribution"][split].append({class_name: result["datasets"][split][class_name] / result["count"][split]})
      result["distribution"][split] = sorted(result["distribution"][split], key=lambda x: list(x.values())[0], reverse=True)

    return result


class EvalObjectDetectionDataset(EvalObjectDetection):
  def __init__(self, predictions, labels, split="test"):
    self.split = split
    super().__init__(predictions.image_labels[split], labels.image_labels[split], predictions.classes)
    self.from_datasets()

  """
    Converted predictions of type [{"class 1": [x1,y1,x2,y2, confidence]}, ...]
    Boxes of type xyxyn + confidence

    Converted Labels of type [{"class 1": [[x1,y1,x2,y2],...], "class 2": [[], [],...]...}]
    Boxes of type xyxyn
  """
  def from_datasets(self):
    self.converted_predictions = []
    self.converted_labels = []

    list_prediction_images = list(self.predictions.keys())
    list_prediction_images.sort()
    list_labels_images = list(self.labels.keys())
    list_labels_images.sort()
    if not all(a == b for a, b in zip(list_prediction_images, list_labels_images)):
      error("predictions and labels differ")
      return

    for image_name in list_prediction_images:
      self.converted_labels.append(self.labels[image_name])
      current_image_prediction = {}
      for class_name, class_prediction in self.predictions[image_name].items():
        current_image_prediction[class_name] = class_prediction
        for i, current_class_prediction in enumerate(current_image_prediction[class_name]):
          current_image_prediction[class_name][i].append(1.0)
      self.converted_predictions.append(current_image_prediction)



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import os

from sciveo.tools.logger import *
from sciveo.common.sampling import RandomSampler, GridSampler, AutoSampler
from sciveo.content.project import RemoteProject, LocalProject


class ProjectRunner:
  current = None

  def __init__(self, project, function, configuration={}, **kwargs):
    self.project_name = project
    self.function = function
    self.kwargs = kwargs

    self.arguments = {
      'count': None,
      'remote': True,
      'sampler': "random",
    }

    self.count = self.get('count')
    remote = self.get('remote')
    sampler = self.get('sampler')

    if remote:
      self.project = RemoteProject(self.project_name)
    else:
      self.project = LocalProject(self.project_name)

    if sampler == "random":
      self.configuration_sampler = RandomSampler(configuration)
    elif sampler == "grid":
      self.configuration_sampler = GridSampler(configuration)
    elif sampler == "auto":
      self.configuration_sampler = AutoSampler(configuration, self.project, **kwargs)
    else:
      self.configuration_sampler = RandomSampler(configuration)

    debug(f"start remote[{remote}] count[{self.count}] sampler[{sampler}]", configuration)

  def get(self, a):
    return self.kwargs.get(a, self.arguments[a])

  def describe(self):
    return {
      "arguments": self.arguments,
      "sampler": self.configuration_sampler.describe()
    }

  def run(self):
    for i, configuration_sample in enumerate(self.configuration_sampler):
      if self.count is not None and i >= self.count:
        break

      self.project.config = configuration_sample
      self.project.config.set_name(f"[{self.project.list_content_size + i + 1}]")
      debug("run", i, self.project.config)
      self.function()



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import boto3

from sciveo.tools.logger import *
from sciveo.media.pipelines.processors.sci.base import *
from sciveo.media.ml.time_series.predictor import *


class TimeSeriesPredictorProcessor(SciBaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)

    self.default.update({
      "time_column": "Date Time",
      "time_format": '%d.%m.%Y %H:%M:%S'
    })

  def run(self, job, input):
    progress_per_media = self.max_progress / max(1, len(input))
    debug("run", job["id"], progress_per_media, "input", input)
    next_media = []
    try:
      if job['content_type'] == "project":
        project_path = os.path.join(self.base_tmp_path, "projects", job['content_id'])
        mkdirs(project_path)

        self.download_content(job, project_path)
        MediaJobState.queue().inc_progress(job["id"], 10)

        list_content = {}

        dataset_df = []
        for content in job['content']:
          list_content.setdefault(content["content_type"], []).append(content)

          if content["content_type"] == "file":
            if content['key'].split(".")[-1] == "timeseries":
              list_content.setdefault("timeseries", []).append(content)

          if content["content_type"] == "dataset":
            # Datasets with S3 located files
            if "bucket" in content and "key" in content:
              remote_path, local_path, local_file_name = self.content_path(content, project_path)
              file_extension = local_file_name.split(".")[-1]

              if file_extension == "csv":
                df = pd.read_csv(local_path)
                x_col = content['data'].get("dataset", {}).get("x_col", self["time_column"])
                df.set_index(x_col, inplace=True)
                dataset_df.append(df)

        dataset_df = pd.concat(dataset_df).drop_duplicates(keep='first').sort_index().reset_index()

        content_project = list_content["project"][0]

        MediaJobState.queue().inc_progress(job["id"], 5)

        if "timeseries" in list_content:
          content_predictor = self.new_content(content_project, "prediction", name=f"Prediction on {dataset_df.shape}")
          next_media.append(content_predictor)

          text_content = self.content_text(content_predictor, "timeseries predictions plots", f"""Time series predictions.\n\n{self.describe_df(dataset_df)}""")
          next_media.append(text_content)

          content = list_content["timeseries"][0]
          remote_path, local_path, local_file_name = self.content_path(content, project_path)
          predictor = TimeSeriesPredictor(local_path)
          ds = TimeSeriesDataSet(dataset_df, self["time_column"], format=self["time_format"])
          predictions, X, x_plot = predictor.predict(ds.data)

          MediaJobState.queue().inc_progress(job["id"], 5)

          columns = predictor.model_data["columns"][:11] # TODO: Should be configurable in the Predictor, so plot only few columns
          plot_progress_inc = 20 / len(columns)
          for i, y_col in enumerate(columns):
            image_content = self.content_image(text_content, y_col, project_path)
            predictor.plot(predictions, X, x_plot, i, image_local_path=image_content["local_path"])
            image_content["data"] = {
              "info": {
                "size": os.path.getsize(image_content["local_path"]),
                "plot": y_col
              }
            }
            next_media.append(image_content)
            MediaJobState.queue().inc_progress(job["id"], plot_progress_inc)
      MediaJobState.queue().inc_progress(job["id"], 10)

      if self["output"]:
        job["output"] += next_media
      if self["append-content"]:
        job["append-content"] += next_media
    except Exception as e:
      exception(e)
    return next_media

  def name(self):
    return "sci-timeseries-predictor"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

from sciveo.tools.logger import *
from sciveo.tools.timers import Timer
from sciveo.tools.configuration import GlobalConfiguration


class Predictors:
  instance = None

  def __init__(self, predictors=None):
    if predictors is not None:
      self.list_predictors = predictors
    else:
      from sciveo.ml.images.embeddings import ImageEmbedding
      from sciveo.ml.nlp.embeddings import TextEmbedding
      self.list_predictors = {
        "TextEmbedding": TextEmbedding(),
        "ImageEmbedding": ImageEmbedding(),
      }

    self.stats = {}

  @staticmethod
  def get():
    if Predictors.instance is None:
      Predictors.instance = Predictors(GlobalConfiguration.get().data.get("API_PREDICTORS", None))
    return Predictors.instance

  @staticmethod
  def set(predictors):
    Predictors.instance = predictors

  def predict(self, predictor_name, X):
    t = Timer()
    predicted = self.list_predictors[predictor_name].predict(X)
    self.stats.setdefault(predictor_name, {"elapsed": 0.0, "count": 0})
    self.stats[predictor_name]["elapsed"] += t.stop()
    self.stats[predictor_name]["count"] += len(X)
    return predicted

  def train(self, predictor_name, X, Y_true=None):
    t = Timer()
    result = self.list_predictors[predictor_name].train(X, Y_true)
    self.stats.setdefault(predictor_name, {"elapsed": 0.0, "count": 0})
    self.stats[predictor_name]["elapsed"] += t.stop()
    self.stats[predictor_name]["count"] += len(X)
    return result

  def describe(self):
    result = []
    for k, v in self.list_predictors.items():
      result.append([k, v.describe(), self.stats.get(k, "")])
    return result

  def print(self):
    for k, v in self.list_predictors.items():
      info("predictor", k, v.describe())



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import numpy as np

from sciveo.tools.logger import *


class BaseOptimizer:
  def __init__(self, keys, use_keys, learning_rate=0.01, learning_rate_decay=0.001):
    # Negative learning rate, Y is loss and will optimize for minimum
    # Positive learning rate, Y is score and optimize for maximium
    self.learning_rate = learning_rate
    self.initial_learning_rate = learning_rate
    self.learning_rate_decay = 1 - learning_rate_decay
    # Gradient based optimisation needs continuous parameters only.
    # So use keys list represents continuous params only. Remain keys list represents the non-continuous params list.
    self.keys = set(keys)
    self.use_keys = set(use_keys)
    self.remain_keys = self.keys - self.use_keys
    self.iteration = 0
    debug("init", self.use_keys, learning_rate, learning_rate_decay)

  def x_to_list(self, x1, x2):
    list_x1 = []
    list_x2 = []
    for k in self.use_keys:
      list_x1.append(x1[k])
      list_x2.append(x2[k])
    return np.array(list_x1), np.array(list_x2)

  def list_to_x(self, x1, x2):
    dict_x = {}
    for k in self.remain_keys:
      dict_x[k] = x1[k]
    for i, k in enumerate(self.use_keys):
      dict_x[k] = x2[i]
    return dict_x

  def gradients(self, x1, y1, x2, y2, max_gradient_norm=1.0):
    delta_x = x1 - x2
    delta_y = y1 - y2
    result = delta_y / (delta_x + 1e-10)
    return np.clip(result, -max_gradient_norm, max_gradient_norm), delta_x, delta_y

  def on_iteration(self):
    self.learning_rate *= self.learning_rate_decay
    self.iteration += 1

  def update(self, x, y, x_prev, y_prev):
    x1, x_prev1 = self.x_to_list(x, x_prev)
    grads, delta_x, delta_y = self.gradients(x1, y, x_prev1, y_prev)

    x_new = x1 * (1 + self.learning_rate * grads)

    self.on_iteration()
    # debug(self.iteration, "update", [x1.tolist(), y], "x_new", x_new.tolist(), "grads", grads.tolist(), "lr", self.learning_rate)

    return delta_y, self.list_to_x(x, x_new)


class HistoryGradientOptimizer(BaseOptimizer):
  def __init__(self, keys, use_keys, learning_rate=0.01, learning_rate_decay=0.001):
    super().__init__(keys, use_keys, learning_rate, learning_rate_decay)

  def update(self, list_samples):
    for i in range(len(list_samples) - 1):
      x1, y1 = list_samples[i]
      x2, y2 = list_samples[i + 1]

      x11, x22 = self.x_to_list(x1, x2)
      grads, delta_x, delta_y = self.gradients(x11, y1, x22, y2)


class AdamOptimizer(BaseOptimizer):
  def __init__(self, keys, use_keys, learning_rate=0.01, learning_rate_decay=0.001, beta1=0.9, beta2=0.999, epsilon=1e-8):
    super().__init__(keys, use_keys, learning_rate, learning_rate_decay)
    self.beta1 = beta1
    self.beta2 = beta2
    self.epsilon = epsilon
    self.t = 0
    self.m = None
    self.v = None

  def update(self, x, y, x_prev, y_prev):
    x1, x_prev1 = self.x_to_list(x, x_prev)
    grads, delta_x, delta_y = self.gradients(x1, y, x_prev1, y_prev)

    if self.m is None:
      self.m = np.zeros_like(x1)
      self.v = np.zeros_like(x1)

    self.t += 1

    self.m = self.beta1 * self.m + (1 - self.beta1) * grads
    self.v = self.beta2 * self.v + (1 - self.beta2) * (grads ** 2)

    m_hat = self.m / (1 - self.beta1 ** self.t)
    v_hat = self.v / (1 - self.beta2 ** self.t)

    x_new = x1 + self.learning_rate * m_hat / (np.sqrt(v_hat) + self.epsilon)

    self.on_iteration()
    # debug(self.iteration, "update", [x1.tolist(), y], "x_new", x_new.tolist(), "grads", grads.tolist(), "lr", self.learning_rate)

    return delta_y, self.list_to_x(x, x_new)



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import random
import string


def random_password(length):
  lower = string.ascii_lowercase
  upper = string.ascii_uppercase
  digits = string.digits
  special = '!@#$%&'

  # Ensure the password contains at least one character from each set
  password = [
    random.choice(lower),
    random.choice(upper),
    random.choice(digits),
    random.choice(special)
  ]

  all_chars = lower + upper + (2 * digits) + (3 * special)
  password += random.choices(all_chars, k=length - 4)
  random.shuffle(password)
  return ''.join(password)



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import cv2

from sciveo.tools.common import *
from sciveo.media.pipelines.processors.base import *


class ImageResize(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.default.update({"heights": [720, 480, 320], "JPEG_QUALITY": 100})

  def process(self, media):
    local_path = media["local_path"]

    frame = cv2.imread(local_path)

    for h in self["heights"]:
      try:
        h, w = get_frame_resolution(frame, h)

        tag = f"resized-{h}"
        resized_local_path = self.add_suffix_to_filename(local_path, tag)

        frame_resized = cv2.resize(frame, (w, h), interpolation=cv2.INTER_AREA)
        cv2.imwrite(resized_local_path, frame_resized, [cv2.IMWRITE_JPEG_QUALITY, self["JPEG_QUALITY"]])

        self.next_content(media, tag, resized_local_path, w=w, h=h)
      except Exception as e:
        exception(e, h)

    return media

  def is_resizer(self):
    return True

  def content_type(self):
    return "image"

  def name(self):
    return "image-resize"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import os
import json
from urllib import request, parse
from urllib.error import HTTPError

from sciveo.tools.logger import *
from sciveo.tools.configuration import GlobalConfiguration


class APIRemoteClient:
  def __init__(self, base_url=None, ver=1):
    self.config = GlobalConfiguration.get()
    self.auth_token = None
    if base_url is None:
      base_url = self.config["api_base_url"]
    self.base_url = f"{base_url}/api/v{ver}/"
    self.headers = { "Auth-Token": self.config['secret_access_key'] }
    debug(f"base url: {self.base_url}")

  def POST_SCI(self, content_type, data, timeout=30):
    return self.POST(f"sci/{content_type}/", data, timeout)

  def POST(self, url, data, timeout=30):
    url = f"{self.base_url}{url}"
    result = False
    try:
      data = parse.urlencode(data).encode("utf-8")
      # debug("POST", url, data)
      resp = request.urlopen(request.Request(url, data=data, headers=self.headers), timeout=timeout)
      result = json.loads(resp.read())
      # debug("POST result", result)
    except HTTPError as e:
      error(e, url, data)
    except Exception as e:
      error(e, url, data)
    return result

  def GET(self, url, timeout=30):
    url = f"{self.base_url}{url}&auth_token={self.config['secret_access_key']}"
    result = False
    try:
      # debug("GET", url)
      resp = request.urlopen(request.Request(url, headers=self.headers), timeout=timeout)
      result = json.loads(resp.read())
      # debug("GET", resp.status, result)
    except HTTPError as e:
      error(e, url)
    except Exception as e:
      error(e, url)
    return result



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import unittest
import json

from sciveo.ml.dataset.object_detection import *


class TestMLDatasets(unittest.TestCase):
  def test_object_detection(self):
    HOME_PATH = os.path.expanduser("~")

    # ds = YOLODataset()
    # ds.load(f"{HOME_PATH}/data/test_yolo")
    # debug("classes", ds.classes)
    # ds.save(f"{HOME_PATH}/data/tmp/yolo-out", classes=["block", "vertical"], copy_images=False)

    ds1 = ObjectDetectionDataset()
    ds1.from_yolo(f"{HOME_PATH}/data/test_yolo")
    stats = ds1.stats()
    debug("stats", stats)
    info("distribution", stats["distribution"])
    # ds.save(f"{HOME_PATH}/data/tmp/od-1", copy_images=False)

    ds2 = ObjectDetectionDataset()
    ds2.from_yolo(f"{HOME_PATH}/data/job-1")
    stats2 = ds2.stats()
    debug("stats", stats2)
    info("distribution", stats2["distribution"])

    evaluator = EvalObjectDetectionDataset(ds1, ds2)
    scores = evaluator.evaluate()
    debug("eval", scores)


if __name__ == '__main__':
  unittest.main()



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import os
import logging
import inspect
from threading import Lock

from sciveo.tools.configuration import GlobalConfiguration


SCIVEO_LOGGER_NAME = "sciveo-log"

_sciveo_global_config = GlobalConfiguration.get()
_sciveo_log_min_level = _sciveo_global_config["SCI_LOG_LEVEL"]
_sciveo_log_lock = Lock()

def _sciveo_get_logger(name):
  logger = logging.getLogger(name)
  if not logger.hasHandlers():
    with _sciveo_log_lock:
      if not logger.hasHandlers():
        log_min_level = logging.getLevelName(_sciveo_log_min_level)
        if (isinstance(log_min_level, str) and log_min_level.startswith("Level")):
          log_min_level = "DEBUG"
        elif isinstance(log_min_level, int) and log_min_level < 10:
          log_min_level = "DEBUG"
        logger.setLevel(log_min_level)

        formatter = logging.Formatter('%(asctime)s [%(thread)d] [%(levelname)s] %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
        ch = logging.StreamHandler()
        ch.setFormatter(formatter)
        logger.addHandler(ch)
        logger.propagate = False
  return logger

def _sciveo_append_classname(*args):
  frame = inspect.currentframe().f_back.f_back
  class_name = frame.f_locals.get('self', None).__class__.__name__ if 'self' in frame.f_locals else None
  if class_name is not None:
    args = (class_name,) + args
  return args

def debug(*args):
  _sciveo_get_logger(SCIVEO_LOGGER_NAME).debug(_sciveo_append_classname(*args))
def info(*args):
  _sciveo_get_logger(SCIVEO_LOGGER_NAME).info(_sciveo_append_classname(*args))
def warning(*args):
  _sciveo_get_logger(SCIVEO_LOGGER_NAME).warning(_sciveo_append_classname(*args))
def error(*args):
  _sciveo_get_logger(SCIVEO_LOGGER_NAME).error(_sciveo_append_classname(*args))
def critical(*args):
  _sciveo_get_logger(SCIVEO_LOGGER_NAME).critical(_sciveo_append_classname(*args))
def exception(e, *args):
  _sciveo_get_logger(SCIVEO_LOGGER_NAME).exception(_sciveo_append_classname(*args))


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import boto3
import math
import numpy as np
import cv2

from sciveo.ml.images.tools import *
from sciveo.ml.base import BaseML
from sciveo.tools.logger import *


"""
Object Detection Bounding Boxes (bbox) of type [x, y, w, h]

If need to use [x1, y1, x2, y2] need to use the inverted convertor bbox_convert_inverted().

IoU between 2 object detections: iou(bbox1, bbox2)

"""

# convert from [x, y, w, h] -> [x1, y1, x2, y2]
def bbox_convert(bbox):
  x1 = bbox[0]
  y1 = bbox[1]
  x2 = x1 + bbox[2]
  y2 = y1 + bbox[3]
  return [x1, y1, x2, y2]

# convert from [x1, y1, x2, y2] -> [x, y, w, h]
def bbox_convert_inverted(bbox):
  x = bbox[0]
  y = bbox[1]
  w = bbox[2] - x
  h = bbox[3] - y
  return [x, y, w, h]

def bbox_norm(bbox, w, h):
  return (bbox[0] / w, bbox[1] / h, bbox[2] / w, bbox[3] / h)

def bbox_denorm(bbox, w, h):
  return (int(bbox[0] * w), int(bbox[1] * h), int(bbox[2] * w), int(bbox[3] * h))

def bbox_center(bbox):
  return (int(bbox[0] + bbox[2] / 2), int(bbox[1] + bbox[3] / 2))

def bbox_area(bbox):
  return bbox[2] * bbox[3]

def iou(bbox1, bbox2):
  x1 = max(bbox1[0], bbox2[0])
  y1 = max(bbox1[1], bbox2[1])
  x2 = min(bbox1[0] + bbox1[2], bbox2[0] + bbox2[2])
  y2 = min(bbox1[1] + bbox1[3], bbox2[1] + bbox2[3])

  if x1 < x2 and y1 < y2:
    a = (x2 - x1) * (y2 - y1)
  else:
    a = 0

  a1 = bbox_area(bbox1)
  a2 = bbox_area(bbox2)
  return a / (a1 + a2 - a)

def bbox_distance(bbox1, bbox2):
  return points_distance(bbox_center(bbox1), bbox_center(bbox2))


"""

Simple Draw object detectios helpers

"""
def image_shape(image):
  return image.shape[1], image.shape[0]

# Draw label bounding boxes of type [x, y, w, h], if [x1, y1, x2, y2] then set convert=False
def draw_label_bboxes(image, bboxes, color, convert=True):
  w, h = image_shape(image)
  for bbox in bboxes:
    if convert:
      bbox = bbox_convert(bbox)
    bbox = bbox_denorm(bbox, w, h)
    cv2.rectangle(image, (bbox[0], bbox[1]), (bbox[2], bbox[3]), color, 2, 1)
  return image


"""

Object Detectors

"""

class ObjectDetectorBase(BaseML):
  def __init__(self, model_path, cache_dir=None, device='cpu', colors=None):
    super().__init__(model_path, cache_dir=cache_dir, device=device)
    self.model_path = model_path

    if colors is None:
      self.colors = [
        (60, 180, 75),    # Green
        (255, 255, 255),   # White
        (245, 130, 48),   # Orange
        (255, 225, 25),   # Yellow
        (0, 130, 200),    # Blue
        (230, 25, 75),    # Red
        (145, 30, 180),   # Purple
        (70, 240, 240),   # Cyan
        (240, 50, 230),   # Magenta
        (210, 245, 60),   # Lime
        (250, 190, 212),  # Pink
        (0, 128, 128),    # Teal
        (220, 190, 255),  # Lavender
        (170, 110, 40),   # Brown
        (128, 0, 0),      # Maroon
        (0, 0, 128),      # Navy
        (128, 128, 0),    # Olive
        (255, 215, 180),  # Peach
        (255, 250, 200),  # Ivory
        (170, 255, 195),  # Mint
      ]
    else:
      self.colors = colors

  def resize(self, image, h):
    ratio = max(image.shape[0], image.shape[1]) / h
    h, w = int(image.shape[0] / ratio), int(image.shape[1] / ratio)
    return cv2.resize(image, (w, h))

  def load(self, image_path):
    image = cv2.imread(image_path)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    return image

  def read_images(self, images_paths):
    X = []
    for image_path in images_paths:
      image = self.load(image_path)
      image = self.resize(image)
      X.append(image)
    return X

  def draw_label_inverted(self, frame, label_text, x_min, y_min, color, font=cv2.FONT_HERSHEY_SIMPLEX, font_scale=0.3, font_thickness=1):
    text_size = cv2.getTextSize(label_text, font, font_scale, font_thickness)[0]
    y_offset = text_size[1] + 4
    x_offset = text_size[0] + 4
    if y_min - y_offset >= 0:
      text_background_top_left = (x_min, y_min - y_offset)
      text_background_bottom_right = (x_min + x_offset, y_min)
    else:
      text_background_top_left = (x_min, y_min)
      text_background_bottom_right = (x_min + x_offset, y_min + y_offset)

    cv2.rectangle(frame, text_background_top_left, text_background_bottom_right, color, cv2.FILLED)
    cv2.putText(frame, label_text, (x_min + 2, text_background_bottom_right[1] - 2), font, font_scale, (0,0,0), font_thickness)

  def draw_object_rectangle_xyxy(self, frame, box, label, color, alpha=0.2, filled=True):
    (x1, y1, x2, y2) = box

    if filled:
      rectangle = np.zeros((y2 - y1, x2 - x1, 3), dtype=np.uint8)
      rectangle[:] = color
      frame[y1:y2, x1:x2] = cv2.addWeighted(rectangle, alpha, frame[y1:y2, x1:x2], 1 - alpha, 0)

    cv2.rectangle(frame, (x1, y1), (x2, y2), color, thickness=1)

    self.draw_label_inverted(frame, label, x1, y1, color=color)

  def draw_object_rectangle_xywh(frame, box, label, color, alpha=0.2, filled=True):
    (x, y, w, h) = box
    (x1, y1, x2, y2) = x, y, x + w, y + h
    self.draw_object_rectangle_xyxy(frame, (x1, y1, x2, y2), label, color, alpha=alpha, filled=filled)


class ObjectDetectorYOLO(ObjectDetectorBase):
  def __init__(self, model_path="yolo11m.pt", cache_dir=None, device='cpu', colors=None):
    super().__init__(model_path, cache_dir=cache_dir, device=device, colors=colors)
    from ultralytics import YOLO
    if self.model_name.startswith("softel"):
      self.model_path = os.path.join(self.cache_dir, self.model_name.replace("/", "---"))
      if os.path.isfile(self.model_path):
        debug(self.model_name, "available", self.model_path)
      else:
        debug("DWN", self.model_name)
        s3 = boto3.client('s3')
        s3.download_file("sciveo-model", self.model_name, self.model_path)

    self.model = YOLO(self.model_path)

  def predict_one(self, x, confidence_threshold=0.5):
    return self.model.predict(x, device=self.device, conf=confidence_threshold, verbose=False)

  def predict(self, X, max_n=64, confidence_threshold=0.5):
    predictions = []

    num_batches = math.ceil(len(X) / max_n)

    for batch_idx in range(num_batches):
      timer = Timer()
      start_idx = batch_idx * max_n
      end_idx = min((batch_idx + 1) * max_n, len(X))

      batch_images = X[start_idx:end_idx]
      batch_predictions = self.model.predict(batch_images, device=self.device, conf=confidence_threshold, verbose=False)
      predictions.extend(batch_predictions)

      elapsed = timer.stop()
      FPS = len(batch_images) / elapsed
      debug(f"batch {batch_idx} / {num_batches}", "elapsed", elapsed, "FPS", FPS, "len", len(batch_images))

    return predictions

  def resize(self, image, h=640):
    return super().resize(image, h)

  def draw(self, image, detections, colors=None):
    if colors is None:
      colors = self.colors
    height, width = image.shape[:2]
    boxes = detections.boxes
    class_names = detections.names
    for i, box in enumerate(boxes):
      bbox = (np.array(box.xyxyn[0]) * np.array([width, height, width, height])).astype(int).tolist()

      confidence = box.conf[0].item()
      class_id = int(box.cls[0].item())
      label = class_names[class_id]
      label_text = f"{label} {int(confidence * 100)}%"

      color = colors[i % len(colors)]
      self.draw_object_rectangle_xyxy(image, bbox, label_text, color)



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import pandas as pd
import tensorflow as tf
from tensorflow.keras.callbacks import Callback

from sciveo.tools.logger import *
from sciveo.media.pipelines.processors.sci.base import *
from sciveo.media.ml.time_series.predictor import *
from sciveo.media.ml.time_series.trainer import *


class TrainerProgressCallback(Callback):
  def __init__(self, job, epochs, max_progress):
    self.job = job
    self.progress_per_epoch = 2 * max_progress / epochs

  def on_epoch_end(self, epoch, logs=None):
    debug("epoch", epoch, "finished")
    MediaJobState.queue().inc_progress(self.job["id"], self.progress_per_epoch)


class TimeSeriesTrainerProcessor(SciBaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)

    self.default.update({
      "model_name": "none",
      "time_column": "Date Time",
      "time_format": '%d.%m.%Y %H:%M:%S',
      "ratios": [["train", 0.70], ["val", 0.20], ["test", 0.10]],
      "max_epochs": 20, "patience": 2,
      "window_size": 24,
      "test_plot_from": 200, "test_plots": 20, "test_plot_column_id": 1
    })

  def run(self, job, input):
    progress_per_media = self.max_progress / max(1, len(input))
    debug("run", job["id"], progress_per_media, "input", input)
    next_media = []
    try:
      if job['content_type'] == "project":
        project_path = os.path.join(self.base_tmp_path, "projects", job['content_id'])
        mkdirs(project_path)

        self.download_content(job, project_path)
        MediaJobState.queue().inc_progress(job["id"], 10)

        list_content = {}

        dataset_df = []
        for content in job['content']:
          list_content.setdefault(content["content_type"], []).append(content)

          if content["content_type"] == "file":
            if content['key'].split(".")[-1] == "timeseries":
              list_content.setdefault("timeseries", []).append(content)

          if content["content_type"] == "dataset":
            # Datasets with S3 located files
            if "bucket" in content and "key" in content:
              remote_path, local_path, local_file_name = self.content_path(content, project_path)
              file_extension = local_file_name.split(".")[-1]

              if file_extension == "csv":
                df = pd.read_csv(local_path)
                x_col = content['data'].get("dataset", {}).get("x_col", df.keys()[0])
                df.set_index(x_col, inplace=True)
                dataset_df.append(df)

        dataset_df = pd.concat(dataset_df).drop_duplicates(keep='first').sort_index().reset_index()

        content_project = list_content["project"][0]
        MediaJobState.queue().inc_progress(job["id"], 5)

        if len(dataset_df) > 0:
          content_trainer = self.new_content(content_project, "training", name=f"Training on {dataset_df.shape}")
          next_media.append(content_trainer)

          ds = TimeSeriesDataSet(dataset_df, self["time_column"], format=self["time_format"])
          ds.normalize()
          ds.split(ratios=self["ratios"])

          trainer = TimeSeriesTrainer(ds, self["window_size"], self["window_size"], self["window_size"])
          trainer.create()
          progress_callback = TrainerProgressCallback(job, self["max_epochs"], 50)
          history = trainer.train(self["max_epochs"], self["patience"], progress_callback)
          trainer_eval = trainer.evaluate()

          model_name, model_path = trainer.save(project_path, self["model_name"])

          model_data = {
            "eval": trainer_eval,
            "history": {
              "loss": history.history["loss"],
              "val_loss": history.history["val_loss"]
            }
          }

          model_content = self.content_file(content_trainer, model_name, model_path, data=model_data)
          next_media.append(model_content)

          text = f"""
            Time series model.
            ....
            ....
            model name {model_name}
            model size {os.path.getsize(model_path)}

            eval
            {self.df_to_html(pd.DataFrame(trainer_eval))}

            train loss
            {self.df_to_html(pd.DataFrame({"loss": history.history["loss"], "val_loss": history.history["val_loss"]}))}
          """
          text_content = self.content_text(content_trainer, "Train timeseries model", text)
          next_media.append(text_content)

          # TODO: Create numeric from history.history["loss"], history.history["val_loss"]

          ds.denormalize("test")
          predictor = TimeSeriesPredictor(model_path)

          # Plot some test predictions
          columns = predictor.model_data["columns"]
          plot_progress_inc = 10 / self["test_plots"]
          y_col = columns[self["test_plot_column_id"]]
          for i in range(self["test_plot_from"], self["test_plot_from"] + self["test_plots"]):
            image_content = self.content_image(text_content, y_col, project_path)
            self.plot_chunk(ds.dataset["test"], predictor, i, self["test_plot_column_id"], image_content["local_path"])
            image_content["data"] = {
              "info": {
                "size": os.path.getsize(image_content["local_path"]),
                "plot": y_col
              }
            }
            next_media.append(image_content)
            MediaJobState.queue().inc_progress(job["id"], plot_progress_inc)

      MediaJobState.queue().inc_progress(job["id"], 10)

      if self["output"]:
        job["output"] += next_media
      if self["append-content"]:
        job["append-content"] += next_media
    except Exception as e:
      exception(e)
    return next_media

  def plot_chunk(self, df, predictor, k, i, image_local_path):
    k += 1
    L = predictor.model_data["window"]["input_width"]
    X = df[-(k + 1)*L:-k*L]

    predictions, X, x_plot = predictor.predict(X)

    idx_from = -k * L
    idx_to = -(k - 1) * L
    if idx_to != 0:
      labels = df[idx_from:idx_to]
    else:
      labels = df[idx_from:]
    labels = labels.values.reshape(X.shape).astype('float32')
    labels = tf.convert_to_tensor(labels)

    predictor.plot(
        predictions, X, x_plot, i, labels,
        image_local_path
    )

  def content_type(self):
    return None

  def name(self):
    return "sci-timeseries-trainer"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import os
import time

import numpy as np

from sciveo.tools.logger import *
from sciveo.api.base import *
from sciveo.common.configuration import Configuration
from sciveo.content.experiment import Experiment, RemoteExperiment
from sciveo.content.dataset import Dataset


class ProjectBase:
  def __init__(self, project_name):
    self.project_name = project_name
    self.guid = None
    self.list_content_size = 0
    self.config = Configuration()

    self.list_experiments = []

    self.project_data = {
      "name": project_name,
      "project": {
        "experiments": [],
        "sort": {
          "experiment": "score"
        },
      }
    }

    debug("init", self.project_name)

  def open(self):
    debug("open", self.project_name)
    self.current_experiment = Experiment(self.project_name, self.guid, self.config)
    return self.current_experiment

  def close(self):
    debug("close", self.project_name)
    self.project_data["project"]["dataset"] = Dataset.get().info
    self.current_experiment.close()
    self.list_experiments.append(self.current_experiment)

  def __enter__(self):
    return self.open()

  def __exit__(self, exc_type, exc_value, traceback):
    self.close()

  def dataset(self, d):
    return Dataset.get(d)


class RemoteProject(ProjectBase):
  def __init__(self, project_name, parent_id=None):
    super().__init__(project_name)

    self.api = APIRemoteClient()

    remote_data = {
      "name": project_name,
      "data": self.project_data
    }
    if parent_id is not None:
      remote_data["parent_id"] = parent_id
    result = self.api.POST_SCI("project", remote_data)
    debug("init api", result)
    if result and "error" not in result:
      if result["name"] == project_name:
        self.guid = result["guid"]
        self.list_content_size = result["list_content_size"]
      else:
        error("Project name mismatch", result, project_name)
    else:
      error("api", remote_data, result)

  def open(self):
    debug("open", self.project_name)
    self.current_experiment = RemoteExperiment(self.project_name, self.guid, self.config)
    return self.current_experiment

  def close(self):
    self.project_data["project"]["experiments"].append(self.current_experiment.name)
    super().close()

    remote_data = {
      "name": self.project_name,
      "guid": self.guid,
      "data": self.project_data
    }
    result = self.api.POST_SCI("project", remote_data)
    debug("close", self.project_name, "api", result)


class LocalProject(ProjectBase):
  def __init__(self, project_name):
    super().__init__(project_name)

  def close(self):
    self.project_data["project"]["experiments"].append(self.current_experiment.data)
    super().close()


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import boto3
import cv2
import io
import base64
from PIL import Image

import torch
from torchvision import models, transforms

import numpy as np
import pandas as pd

from transformers import AutoTokenizer, AutoModel

from sciveo.tools.logger import *
from sciveo.ml.images.base import BaseImageML


class ImageEmbedding(BaseImageML):
  def __init__(self, model_id=1, cache_dir=None, device=None) -> None:
    super().__init__(model_id, cache_dir=cache_dir, device=device)

  def init_models(self):
    self.model_name = [
      "softel-resnet18-embedding.pth",
      "softel-resnet34-embedding.pth",
      "softel-resnet50-embedding.pth",
      "softel-resnet101-embedding.pth",
      "softel-resnet152-embedding.pth",
    ][int(self.model_name)]
    self.model_path = os.path.join(self.cache_dir, self.model_name)
    if os.path.isfile(self.model_path):
      debug(self.model_name, "available", self.model_path)
    else:
      debug("DWN", self.model_name)
      s3 = boto3.client('s3')
      s3.download_file("sciveo-model", self.model_name, self.model_path)

    self.preprocessor = transforms.Compose([
      transforms.Lambda(lambda img: img.convert("RGB") if img.mode != "RGB" else img),
      transforms.Resize(256),
      transforms.CenterCrop(224),
      transforms.ToTensor(),
      transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
    ])

    self.model = None

  def load_model(self):
    debug("loading model", self.model_name, self.model_path)
    self.model = torch.load(self.model_path).to(self.device)
    self.model.eval()

  def init(self):
    if self.model is None:
      self.load_model()
      self.post_init()

  def embed(self, image):
    image_tensor = self.preprocessor(image).unsqueeze(0).to(self.device)
    with torch.no_grad():
      embedding = self.model(image_tensor).to("cpu")
    return embedding.squeeze().numpy()

  def read_image(self, x):
    if isinstance(x, dict):
      if "local_path" in x:
        return Image.open(x["local_path"])
      elif "guid" in x:
        debug("not implemented guid", x["guid"])
      else:
        debug("not implemented", x)
    else:
      return Image.open(io.BytesIO(base64.b64decode(x)))

  # TODO: should conform to the BaseImageML.predict_one() on image input
  def predict_one(self, x):
    self.init()
    image = self.read_image(x)
    return self.embed(image)

  def predict(self, X):
    self.init()
    predictions = []
    for current_x in X:
      embedding = self.predict_one(current_x)
      if not isinstance(embedding, list):
        embedding = embedding.tolist()
      predictions.append(embedding)
    return predictions



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2025
#


import psutil
import subprocess
import time
import threading

from sciveo.tools.daemon import *
from sciveo.tools.logger import *
from sciveo.tools.simple_counter import RunCounter


class MemoryWatchDogDaemon(DaemonBase):
  def __init__(self, threshold_percent=90, period=5, command="echo '⚠️ Low Memory!'"):
    super().__init__(period=period)
    self.threshold_percent = threshold_percent
    self.command = command
    self.used_percent = 0
    self.printer = RunCounter(60, lambda: debug(f"Memory usage: {self.used_percent}%", f"threshold: {self.threshold_percent}%"))

  def loop(self):
    mem = psutil.virtual_memory()
    self.used_percent = mem.percent
    self.printer.run()
    if self.used_percent > self.threshold_percent:
      warning(f"⚠️ Memory usage {self.used_percent}% exceeded {self.threshold_percent}%, executing command: {self.command}")
      subprocess.run(self.command, shell=True)



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import cv2
import gc
from PIL import Image
import numpy as np

import torch
from transformers import AutoImageProcessor, AutoModelForDepthEstimation

from sciveo.tools.logger import *
from sciveo.tools.os import replace_extension_to_filename
from sciveo.tools.common import *
from sciveo.media.pipelines.processors.tpu_base import *


class ImageDepthEstimation(TPUBaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.default.update({
      "JPEG_QUALITY": 80, "height": 720,
      "model_ver": 0.9
    })

    TPU = os.environ.get("MEDIA_PROCESSING_BACKEND", "cpu")

    model_name =  f"softel/depth-anything-v{self['model_ver']}"
    cache_dir = os.path.join(os.environ['MEDIA_MODELS_BASE_PATH'], "models/")
    self.image_processor = AutoImageProcessor.from_pretrained(
      model_name,
      cache_dir=cache_dir, resume_download=True
    )
    self.model = AutoModelForDepthEstimation.from_pretrained(
      model_name,
      cache_dir=cache_dir, resume_download=True
    ).to(TPU)

  def process(self, media):
    try:
      self.media = media
      self.local_path = media["local_path"]

      tag = "DEPTH"
      image_local_path = self.add_suffix_to_filename(self.local_path, tag)
      image_local_path = replace_extension_to_filename(image_local_path, "jpg")

      image = cv2.cvtColor(cv2.imread(self.local_path), cv2.COLOR_BGR2RGB)
      h, w, c = image.shape
      h, w = get_resize_resoulution(h, w, self["height"])

      inputs = self.image_processor(images=image, return_tensors="pt")
      with torch.no_grad():
        outputs = self.model(**inputs)
        predicted_depth = outputs.predicted_depth
      prediction = torch.nn.functional.interpolate(
        predicted_depth.unsqueeze(1),
        size=(h, w),
        mode="bicubic",
        align_corners=False,
      )
      output = prediction.squeeze().cpu().numpy()
      formatted = (output * 255 / np.max(output)).astype("uint8")

      image_depth = cv2.applyColorMap(formatted, cv2.COLORMAP_JET)
      cv2.imwrite(image_local_path, image_depth, [cv2.IMWRITE_JPEG_QUALITY, self["JPEG_QUALITY"]])

      self.next_content(self.media, tag=tag, local_path=image_local_path, w=w, h=h)
    except Exception as e:
      exception(e, self.media)
    return self.media

  def content_type(self):
    return "image"

  def name(self):
    return "image-depth-estimation"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import numpy as np
import cv2


class BaseImageTransformer:
  def __init__(self, param=0):
    self.param = param

  def __call__(self, image, labels=None):
    return image, labels

  def transform(self, image, labels=None):
    return self(image, labels)


class RandomTransformer(BaseImageTransformer):
  def __init__(self, class_name, lower, upper, proba):
    self.class_name = class_name
    self.lower = lower
    self.upper = upper
    self.proba = proba
    self.transformer = class_name(1.0)

  def __call__(self, image, labels=None):
    p = np.random.uniform(0, 1)
    if p >= (1.0 - self.proba):
      self.transformer.param = np.random.uniform(self.lower, self.upper)
      return self.transformer(image, labels)
    else:
      return image, labels


class ConvertColor(BaseImageTransformer):
  def __init__(self, current='RGB', to='HSV', keep_3ch=True):
    if not ((current in {'RGB', 'HSV'}) and (to in {'RGB', 'HSV', 'GRAY'})): raise NotImplementedError
    self.current = current
    self.to = to
    self.keep_3ch = keep_3ch

  def __call__(self, image, labels=None):
    if self.current == 'RGB' and self.to == 'HSV':
      image = cv2.cvtColor(image, cv2.COLOR_RGB2HSV)
    elif self.current == 'RGB' and self.to == 'GRAY':
      image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
      if self.keep_3ch:
        image = np.stack([image] * 3, axis=-1)
    elif self.current == 'HSV' and self.to == 'RGB':
      image = cv2.cvtColor(image, cv2.COLOR_HSV2RGB)
    elif self.current == 'HSV' and self.to == 'GRAY':
      image = cv2.cvtColor(image, cv2.COLOR_HSV2GRAY)
      if self.keep_3ch:
        image = np.stack([image] * 3, axis=-1)
    return image, labels


class ConvertDataType(BaseImageTransformer):
  def __init__(self, to=np.uint8):
    self.to = to

  def __call__(self, image, labels=None):
    if self.to == np.uint8:
      image = np.round(image, decimals=0).astype(self.to)
    else:
      image = image.astype(np.float32)
    return image, labels


class ConvertTo3Channels(BaseImageTransformer):
  def __init__(self):
    pass

  def __call__(self, image, labels=None):
    if image.ndim == 2:
      image = np.stack([image] * 3, axis=-1)
    elif image.ndim == 3:
      if image.shape[2] == 1:
        image = np.concatenate([image] * 3, axis=-1)
      elif image.shape[2] == 4:
        image = image[:,:,:3]
    return image, labels


class Hue(BaseImageTransformer):
  def __init__(self, delta):
    if not (-180 <= delta <= 180):
      delta = 0
    super().__init__(delta)

  def __call__(self, image, labels=None):
    image[:, :, 0] = (image[:, :, 0] + self.param) % 180.0
    return image, labels


class Saturation(BaseImageTransformer):
  def __init__(self, factor):
    if factor <= 0.0:
      factor = 1e-10
    super().__init__(factor)

  def __call__(self, image, labels=None):
    image[:,:,1] = np.clip(image[:,:,1] * self.param, 0, 255)
    return image, labels


class Brightness(BaseImageTransformer):
  def __init__(self, delta):
    super().__init__(delta)

  def __call__(self, image, labels=None):
    image = np.clip(image + self.param, 0, 255)
    return image, labels


class Contrast(BaseImageTransformer):
  def __init__(self, factor):
    if factor <= 0.0: raise ValueError("factor <= 0.0")
    super().__init__(factor)

  def __call__(self, image, labels=None):
    image = np.clip(127.5 + self.param * (image - 127.5), 0, 255)
    return image, labels


class Gamma(BaseImageTransformer):
  def __init__(self, gamma):
    if gamma <= 0.0: raise ValueError("gamma <= 0.0")
    self.gamma = gamma
    self.gamma_inv = 1.0 / gamma
    self.lut = np.array([((i / 255.0) ** self.gamma_inv) * 255 for i in np.arange(0, 256)]).astype("uint8")

  def __call__(self, image, labels=None):
    image = cv2.LUT(image, self.lut)
    return image, labels


class RandomGamma(BaseImageTransformer):
  def __init__(self, lower=0.25, upper=2.0, prob=0.5):
    if lower >= upper: raise ValueError("lower >= upper")
    self.lower = lower
    self.upper = upper
    self.prob = prob

  def __call__(self, image, labels=None):
    p = np.random.uniform(0,1)
    if p >= (1.0-self.prob):
      gamma = np.random.uniform(self.lower, self.upper)
      change_gamma = Gamma(gamma=gamma)
      return change_gamma(image, labels)
    else:
      return image, labels


class HistogramEqualization(BaseImageTransformer):
  def __init__(self):
    pass

  def __call__(self, image, labels=None):
    image[:,:,2] = cv2.equalizeHist(image[:,:,2])
    return image, labels


class RandomHistogramEqualization(BaseImageTransformer):
  def __init__(self, prob=0.5):
    self.prob = prob
    self.equalize = HistogramEqualization()

  def __call__(self, image, labels=None):
    p = np.random.uniform(0,1)
    if p >= (1.0-self.prob):
      return self.equalize(image, labels)
    else:
      return image, labels


class ChannelSwap(BaseImageTransformer):
  def __init__(self, order):
    self.order = order

  def __call__(self, image, labels=None):
    image = image[:,:,self.order]
    return image, labels


class RandomChannelSwap(BaseImageTransformer):
  def __init__(self, prob=0.5):
    self.prob = prob
    self.permutations = [(0, 2, 1), (1, 0, 2), (1, 2, 0), (2, 0, 1), (2, 1, 0)]
    self.swap_channels = ChannelSwap(order=(0, 1, 2))

  def __call__(self, image, labels=None):
    p = np.random.uniform(0,1)
    if p >= (1.0-self.prob):
      i = np.random.randint(5)
      self.swap_channels.order = self.permutations[i]
      return self.swap_channels(image, labels)
    else:
      return image, labels


class BlackWhite(BaseImageTransformer):
  def __init__(self, param=0):
    super().__init__(param)

  def __call__(self, image, labels=None):
    image = np.transpose(np.tile(cv2.cvtColor(image, cv2.COLOR_BGR2GRAY), (3, 1, 1)), [1, 2, 0])
    return image, labels


class GaussianNoise(BaseImageTransformer):
  def __init__(self, variance=64, mean=0):
    super().__init__(variance)
    self.mean = mean

  def __call__(self, image, labels=None):
    if image.ndim < 3:
      image = np.expand_dims(image, axis=-1)
    image = image.astype(np.float32)

    sigma = self.param ** 0.5
    noise = np.random.normal(self.mean, sigma, (image.shape[0], image.shape[1]))
    noise = np.transpose(np.tile(noise, (image.shape[-1], 1, 1)), [1, 2, 0])

    image += noise

    image = cv2.normalize(image, None, 0, 255, cv2.NORM_MINMAX, dtype=-1)
    image = image.astype(np.uint8)

    return image, labels


class Pixelisator(BaseImageTransformer):
  def __init__(self, k):
    super().__init__(k)

  def __call__(self, image, labels=None):
    w = image.shape[0]
    h = image.shape[1]
    k = 1.0 / self.param
    image = cv2.resize(image, (int(w * k), int(h * k)))
    image = cv2.resize(image, (w, h))
    return image, labels


class Blur(BaseImageTransformer):
  def __init__(self, k):
    super().__init__(k)

  def __call__(self, image, labels=None):
    image = cv2.blur(image, (int(self.param), int(self.param)))
    return image, labels


class PhotometricTransformer:
  def __init__(self):
    self.sequences = [
                      [
                        ConvertTo3Channels(),
                        RandomTransformer(Brightness, lower=-32, upper=32, proba=0.5),
                        ConvertDataType(to=np.float32),
                        RandomTransformer(Contrast, lower=0.5, upper=1.5, proba=0.5),
                        ConvertDataType(to=np.uint8),
                        RandomTransformer(BlackWhite, lower=0, upper=0, proba=0.2),
                        RandomTransformer(GaussianNoise, lower=32, upper=256, proba=0.2),
                        RandomTransformer(Pixelisator, lower=1.0, upper=3.0, proba=0.2),
                        RandomTransformer(Blur, lower=2.0, upper=10.0, proba=0.2),
                        ConvertColor(current='RGB', to='HSV'),
                        ConvertDataType(to=np.float32),
                        RandomTransformer(Saturation, lower=0.5, upper=1.5, proba=0.5),
                        RandomTransformer(Hue, lower=-18, upper=18, proba=0.5),
                        ConvertDataType(to=np.uint8),
                        ConvertColor(current='HSV', to='RGB'),
                        RandomChannelSwap(prob=0.0)
                      ],
                      [
                        ConvertTo3Channels(),
                        ConvertDataType(to=np.uint8),
                        RandomTransformer(Brightness, lower=-64, upper=64, proba=0.5),
                        RandomTransformer(BlackWhite, lower=0, upper=0, proba=0.4),
                        RandomTransformer(GaussianNoise, lower=32, upper=256, proba=0.4),
                        RandomTransformer(Pixelisator, lower=1.0, upper=4.0, proba=0.2),
                        RandomTransformer(Blur, lower=2.0, upper=20.0, proba=0.2),
                        ConvertColor(current='RGB', to='HSV'),
                        ConvertDataType(to=np.float32),
                        RandomTransformer(Saturation, lower=0.5, upper=2.5, proba=0.5),
                        RandomTransformer(Hue, lower=-64, upper=64, proba=0.5),
                        ConvertDataType(to=np.uint8),
                        ConvertColor(current='HSV', to='RGB'),
                        ConvertDataType(to=np.float32),
                        RandomTransformer(Contrast, lower=0.5, upper=2.5, proba=0.5),
                        ConvertDataType(to=np.uint8),
                        RandomChannelSwap(prob=0.0)
                      ]
    ]

  def load_image(self, image_path):
    return cv2.imread(image_path)

  def __call__(self, image, labels=None):
    if isinstance(image, str):
      image = self.load_image(image)

    idx = np.random.choice(len(self.sequences))
    for transformer in self.sequences[idx]:
      image, labels = transformer(image, labels)
    return image, labels



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import time
import os
import boto3

from sciveo.tools.logger import *
from sciveo.tools.daemon import *
from sciveo.tools.synchronized import ListQueue
from sciveo.tools.os import *
from sciveo.tools.http import *
from sciveo.media.pipelines.pipeline import *
from sciveo.media.pipelines.queues import MediaJobState


class JobQueueDaemon(DaemonBase):
  _queue = None

  @staticmethod
  def queue():
    if JobQueueDaemon._queue is None:
      JobQueueDaemon._queue = ListQueue("pipeline jobs")
    return JobQueueDaemon._queue

  def __init__(self, base_tmp_path):
    super().__init__()
    self.s3 = boto3.client('s3')
    self.base_tmp_path = base_tmp_path

  def run(self):
    while(self.is_running):
      try:
        job = JobQueueDaemon.queue().pop()
        job = self.process_job(job)
      except Exception as e:
        exception(e, type(self).__name__)
        time.sleep(3)

  def process_job(self, job):
    debug("process_job job", job)
    job["input"] = job["content"]
    MediaJobState.queue().start(job["id"])
    MediaJobState.queue().inc_progress(job["id"], 10)

    pipeline = MediaPipeline(job)
    job = pipeline.run()

    MediaJobState.queue().finish(job["id"], job)
    # debug("job output", job["output"])
    return job



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import time
import threading
import cv2
import numpy as np
from scapy.all import sniff, IP, TCP, UDP

from sciveo.tools.logger import *
from sciveo.tools.timers import FPSCounter
from sciveo.network.tools import StreamSniffer


class RTSPStreamSniffer(StreamSniffer):
  def __init__(self, iface=None, port=554):
    super().__init__(iface)
    self.port = port

  def on_packet(self, packet):
    if self.is_rtsp_packet(packet):
      self.append_ip_packet(packet)

  def is_rtsp_packet(self, packet):
    return IP in packet and TCP in packet and (packet[TCP].dport == self.port or packet[TCP].sport == self.port)
    # return IP in packet and ( (TCP in packet and (packet[TCP].dport == self.port or packet[TCP].sport == self.port)) or (UDP in packet and (packet[UDP].dport == self.port or packet[UDP].sport == self.port)) )

  def get_rtsp_frames(self, ip_src):
    frames = []
    current_packets = self.get_ip_stream(ip_src)
    for packet in current_packets:
      frame = self.packet_to_frame(packet)
      if frame is not None:
        frames.append(frame)
    return frames

  def packet_to_frame(self, packet):
    payload = bytes(packet[TCP].payload)
    nparr = np.frombuffer(payload, np.uint8)
    frame = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
    return frame

  def play(self, ip_src):
    fps = FPSCounter(tag=f"play {ip_src}")
    while True:
      frames = self.get_rtsp_frames(ip_src)
      for frame in frames:
        fps.update()
        cv2.imshow(f'RTSP Stream from {ip_src}', frame)
        if cv2.waitKey(1) & 0xFF == ord('q'):
          return

  def play_cams(self, list_cams):
    threads = []
    for ip in list_cams:
      t = threading.Thread(target=self.play, args=(ip,))
      t.start()
      threads.append(t)

    self.start()

    try:
      while True:
        time.sleep(1)
    except KeyboardInterrupt:
      self.stop()
      for t in threads:
        t.join()
      cv2.destroyAllWindows()


if __name__ == '__main__':
  sniffer = RTSPStreamSniffer(iface="en0", port=554)
  sniffer.play_cams([])



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import cv2
import re
import datetime

from PIL import Image
from PIL.ExifTags import TAGS, GPSTAGS

from sciveo.tools.logger import *
from sciveo.tools.common import *
from sciveo.media.pipelines.processors.base import *
from sciveo.media.pipelines.queues import MediaJobState
from sciveo.media.pipelines.base import ApiContent


class MediaInfo(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.api = ApiContent()
    self.default = {}

  def update(self, media, info):
    data = {"data": {"info": info}}

    if "datetime" in info:
      try:
        data["created_at"] = datetime.datetime.strptime(info["datetime"], '%Y:%m:%d %H:%M:%S')
      except Exception:
        pass

    debug("update", media, data)
    self.api.update(media, data)

  def fix_value(self, value):
    if isinstance(value, bytes):
      try:
        value = value.decode('utf-8')
      except Exception:
        value = ""
    if isinstance(value, str):
      value = re.sub(r'[\x00-\x1f\x7f-\xff]', '', value)
    if isinstance(value, tuple):
      value = list(value)
    if isinstance(value, list):
      result = []
      for v in value:
        result.append(self.fix_value(v))
      value = result
    if isinstance(value, dict):
      result = {}
      for k, v in value.items():
        result[k] = self.fix_value(v)
      value = result
    return value

  def run(self, job, input):
    if len(input) == 0:
      return []

    images = []
    videos = []

    for i, media in enumerate(input):
      if media["content_type"] == "image":
        images.append(media)
      if media["content_type"] == "video":
        videos.append(media)

    progress_per_media = self.max_progress / (len(images) + len(videos))
    debug("run", job["id"], progress_per_media)

    for media in images:
      try:
        info = { "size": os.path.getsize(media["local_path"]) }

        with Image.open(media["local_path"]) as img:
          info["width"] = img.width
          info["height"] = img.height
          info["format"] = img.format
          info["layers"] = img.layers
          info["mode"] = img.mode

          if "dpi" in img.info:
            info["dpi"] = list(img.info["dpi"])

          exif = img._getexif()
          if exif:
            info["exif"] = {}
            for tag, value in exif.items():
              decoded_tag = str(TAGS.get(tag, tag))

              if decoded_tag == "GPSInfo":
                gps_data = {}
                for t, v in value.items():
                  sub_decoded = GPSTAGS.get(t, t)
                  gps_data[sub_decoded] = v
                value = gps_data

              value = self.fix_value(value)

              info["exif"][decoded_tag] = value

            for k in ["DateTime", "DateTimeOriginal", "DateTimeDigitized"]:
              if k in info["exif"]:
                info["datetime"] = info["exif"][k]
                break

        self.update(media, info)
        MediaJobState.queue().inc_progress(job["id"], progress_per_media)
      except Exception as e:
        exception(e)

    for media in videos:
      try:
        info = { "size": os.path.getsize(media["local_path"]) }

        cap = cv2.VideoCapture(media["local_path"])
        frames_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        fps = int(cap.get(cv2.CAP_PROP_FPS))

        info["frames_count"] = frames_count
        info["fps"] = fps
        info["vls"] = frames_count / fps

        ret, frame = cap.read()
        if ret:
          info["width"] = frame.shape[1]
          info["height"] = frame.shape[0]

        cap.release()

        self.update(media, info)
        MediaJobState.queue().inc_progress(job["id"], progress_per_media)
      except Exception as e:
        exception(e)

    return []

  def content_type(self):
    return "media"

  def name(self):
    return "media-info"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

import cv2

from sciveo.tools.common import *
from sciveo.media.pipelines.processors.base import *


class ImageHistogram(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.default.update({"width": 640, "height": 480})

  def draw(self, frame, image_local_path, w, h):
    dpi = 100
    fig = plt.figure(figsize=(int(w)/dpi, int(h)/dpi), dpi=dpi)

    ax1 = fig.add_subplot(2, 1, 1)
    ax2 = fig.add_subplot(2, 1, 2)

    color = ('b','g','r')
    for i,col in enumerate(color):
      ax1.plot(cv2.calcHist([frame], [i], None, [256], [0, 258]), color=col)

    gray_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    # ax2.plot(cv2.calcHist([gray_frame], [0], None, [256], [0, 256]))
    ax2.hist(gray_frame.ravel(), 256, [0, 258])

    ax1.set_title("RGB Histogram")
    ax1.axis(xmin=0, xmax=256)
    ax2.set_title("Gray Histogram")
    ax2.axis(xmin=0, xmax=258)

    plt.subplots_adjust(hspace=0.5)
    plt.savefig(image_local_path, dpi=dpi, bbox_inches='tight', pad_inches=0)

  def process(self, media):
    try:
      self.media = media
      self.local_path = media["local_path"]

      tag = "HIST"
      image_local_path = self.add_suffix_to_filename(self.local_path, tag)

      w, h = self["width"], self["height"]

      frame = cv2.imread(self.local_path)
      self.draw(frame, image_local_path, w, h)
      self.next_content(self.media, tag, image_local_path, w=w, h=h)

    except Exception as e:
      exception(e, self.media)
    return self.media

  def content_type(self):
    return "image"

  def name(self):
    return "image-histogram"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import math
import numpy as np


def points_distance(p1, p2):
  p1 = np.array(p1)
  p2 = np.array(p2)
  return np.linalg.norm(p1 - p2)

def validate_point(p, w, h):
  x = max(p[0], 0)
  x = min(x, w)
  y = max(p[1], 0)
  y = min(y, h)
  return (x, y)

# Calc diameter point from a circle point and circle center
def diameter_point(p1, p2):
  x = 2 * p2[0] - p1[0]
  y = 2 * p2[1] - p1[1]
  return (x, y)

# Center point for a line section (with some ratio for general purpose)
def line_section_center_point(p1, p2, ratio=0.5):
  x = int(p1[0] + ratio * (p2[0] - p1[0]))
  y = int(p1[1] + ratio * (p2[1] - p1[1]))
  return (x, y)

# Line equation ax + by + c = 0 coefficients
def line_coeff(l):
  a = l[0][1] - l[1][1]
  b = l[1][0] - l[0][0]
  c = l[0][0] * l[1][1] - l[1][0] * l[0][1]
  return a, b, c

# 2 Lines intersection point (None, None) when parallel
def lines_intersection(l1, l2):
  a1, b1, c1 = line_coeff(l1)
  a2, b2, c2 = line_coeff(l2)

  d = (a1 * b2 - a2 * b1)
  if d != 0.0:
    x = (b1 * c2 - b2 * c1) / d
    y = (c1 * a2 - c2 * a1) / d
  else:
    x, y = None, None
  return x, y

# Check if p is between p1 and p2 on a line
def line_point_between(p, p1, p2):
  return (
        p[0] >= min(p1[0], p2[0]) and p[0] <= max(p1[0], p2[0])
    and p[1] >= min(p1[1], p2[1]) and p[1] <= max(p1[1], p2[1])
  )

# Check a point p is intersection for 2 line segments l1 and l2
def is_point_segments_intersection(p, l1, l2):
  return line_point_between(p, l1[0], l1[1]) and line_point_between(p, l2[0], l2[1])

# Angle from radians to degrees convertor
def angle_rad2deg(angle):
  return math.fabs(angle * 180 / math.pi)

# Angle between lines in radians
def lines_angle(l1, l2):
  v1 = (l1[0][0] - l1[1][0], l1[0][1] - l1[1][1])
  v2 = (l2[0][0] - l2[1][0], l2[0][1] - l2[1][1])
  return math.atan2(v1[0], v1[1]) - math.atan2(v2[0], v2[1])

# Angle between lines in degrees
def lines_angle_deg(l1, l2):
  return angle_rad2deg(lines_angle(l1, l2))



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

from sciveo.media.ml.encoders.base import *


class Normalizer:
  def __init(self):
    self.mean = 0
    self.std = 0

  def fit(self, X):
    self.mean = X.mean(axis=0)
    self.std = X.std(axis=0)

  def transform(self, X):
    return (X - self.mean) / self.std

  def fit_transform(self, X):
    self.fit(X)
    return self.transform(X)

  def inverse(self, X):
    return X * self.std + self.mean


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import qrcode
from PIL import Image

from sciveo.media.pipelines.processors.base import *


class QRGenerator(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.default.update({
      "text": "https://smiveo.com",
      "width": 512,
      "height": 512,
      "bucket": "smiveo-images"
    })

  def run(self, job, input):
    width = self["width"]
    height = self["height"]

    list_frames = []

    parent_guid = job['content_id']
    guid = self.new_guid()
    key = f"{guid}.jpg"
    bucket = self["bucket"]

    image_local_path = os.path.join(self.base_tmp_path, key)

    qr = qrcode.QRCode(
      version=1,
      error_correction=qrcode.constants.ERROR_CORRECT_L,
      box_size=10,
      border=4,
    )
    qr.add_data(self["text"])
    qr.make(fit=True)
    qr_image = qr.make_image(fill_color="black", back_color="white")
    qr_image.save(image_local_path)

    media = {
      "guid": guid,
      "parent": parent_guid,
      "content_type": "image",
      "owner": job["owner"],
      "name": f"QR {self['text'][:8]}",
      "description": f"QR [{self['text']}]",
      "local_path": image_local_path,
      "w": width, "h": height,
      "height": height,
      "key": key,
      "bucket": bucket,
      "processor": self.name(),
      "layout": {"name": self.name(), "height": height, **self["layout"]}
    }

    if self["output"]:
      job["output"].append(media)
    if self["append-content"]:
      job["append-content"].append(media)
    return [media]

  def name(self):
    return "QR-generator"

  def is_append_processor(self):
    return False


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import matplotlib.pyplot as plt
import matplotlib.dates as mdates

import numpy as np
import pandas as pd

import tensorflow as tf
from tensorflow import keras

from sciveo.tools.logger import *
from sciveo.media.ml.encoders.normalizer import *


class WindowGenerator():
  def __init__(self, input_width, label_width, shift):
    self.input_width = input_width
    self.label_width = label_width
    self.shift = shift

    self.total_window_size = input_width + shift

    self.input_slice = slice(0, input_width)
    self.input_indices = np.arange(self.total_window_size)[self.input_slice]

    self.label_start = self.total_window_size - self.label_width
    self.labels_slice = slice(self.label_start, None)
    self.label_indices = np.arange(self.total_window_size)[self.labels_slice]

  def split_window(self, features):
    X = features[:, self.input_slice, :]
    Y_label = features[:, self.labels_slice, :]

    X.set_shape([None, self.input_width, None])
    Y_label.set_shape([None, self.label_width, None])

    return X, Y_label

  def transform(self, data):
    data = np.array(data, dtype=np.float32)
    ds = tf.keras.utils.timeseries_dataset_from_array(
      data=data,
      targets=None,
      sequence_length=self.total_window_size,
      sequence_stride=1,
      shuffle=True,
      batch_size=32,
    )
    ds = ds.map(self.split_window)
    return ds

  def __repr__(self):
    return '\n'.join([
        f'Total window size: {self.total_window_size}',
        f'Input indices: {self.input_indices}',
        f'Label indices: {self.label_indices}',
    ])

  def data(self, df=None, normalizer=None):
    result = getattr(self, 'data_', None)
    if result is None and df is not None:
      result = next(iter(self.transform(df)))
      self.data_ = result
    if normalizer is not None:
      self.normalizer_ = normalizer
    return result

  def predictions(self, model, X=None, Y_label=None):
    if X is None or Y_label is None:
      X, Y_label = self.data()
    self.predictions_ = model(X)

  def plot(self, plot_col, plot_col_index, X=None, Y_label=None, predictions=None, normalizer=None, max_subplots=10):
    if normalizer is None:
      normalizer = self.normalizer_
    if X is None or Y_label is None:
      X, Y_label = self.data()
    if predictions is None and getattr(self, 'predictions_', None) is not None:
      predictions = self.predictions_

    X_denorm = normalizer.inverse(X)
    Y_label_denorm = normalizer.inverse(Y_label)

    plt.figure(figsize=(12, 8))
    max_n = min(max_subplots, len(X))
    for n in range(max_n):
      plt.subplot(max_n, 1, n+1)
      plt.ylabel(f'{plot_col}')

      plt.plot(
        self.input_indices, X_denorm[n, :, plot_col_index],
        label='X', marker='.', zorder=-10
      )

      plt.scatter(
        self.label_indices, Y_label_denorm[n, :, plot_col_index],
        edgecolors='k', label='Y_label', c='#2ca02c', s=64
      )

      if predictions is not None:
        predictions_denorm = normalizer.inverse(predictions)
        plt.scatter(
          self.label_indices, predictions_denorm[n, :, plot_col_index],
          marker='X', edgecolors='k', label='Predictions',
          c='#ff7f0e', s=64
        )

      if n == 0:
        plt.legend()

    plt.xlabel("time")



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import time
import json

from fastapi import FastAPI, Query, Request
from fastapi.responses import HTMLResponse
from typing import Dict, Any, Optional, Union
from pydantic import BaseModel
import uvicorn

from sciveo.tools.logger import *
from sciveo.tools.daemon import *
from sciveo.media.pipelines.job_daemon import *
from sciveo.media.pipelines.pipeline import *
from sciveo.media.pipelines.queues import MediaJobState


def check_auth(auth_key):
  if auth_key != os.environ['AUTH_KEY_MEDIA_PIPELINES_WEB']:
    raise Exception("no access, auth key invalid")

class JobData(BaseModel):
  auth_token: str
  data: Optional[Dict]


app = FastAPI()

@app.get("/")
async def index():
  return HTMLResponse(
      """
        <html>
          <head>
            <title>SMIVEO</title>
          </head>
          <body>
            <div>SMIVEO</div>
          </body>
        </html>
      """
  )

@app.get("/job/{job_id}")
async def job_info(job_id: int, auth_token: str, data: Union[str, None] = None):
  try:
    check_auth(auth_token)
    data = json.loads(data)
    return {"job": job_id, "data": data}
  except Exception as e:
    exception(e, "job_info")
    return {"error": str(e)}

@app.post("/job")
async def job_start(auth_token: str, job: Union[Dict, None] = None):
  try:
    check_auth(auth_token)
    debug("job_start", job.data)
    JobQueueDaemon.queue.push(job.data)
    return {"job": job.data}
  except Exception as e:
    exception(e, "job_start")
    return {"error": str(e)}


class WebMediaPipelinesDaemon(DaemonBase):
  def __init__(self, host="0.0.0.0", port=8001):
    super().__init__()
    self.host = host
    self.port = port

  def run(self):
    info("FastAPI", self.host, self.port)
    uvicorn.run(app, host=self.host, port=self.port, log_level="warning")

  def info(self):
    return "web media pipelines (FastAPI )"


if __name__ == "__main__":

  daemons = [
    JobQueueDaemon(os.environ["WEB_MEDIA_PIPELINES_TMP_PATH"]),
    WebMediaPipelinesDaemon(port=os.environ["WEB_MEDIA_PIPELINES_PORT"])
  ]

  for daemon in daemons:
    debug("start", type(daemon).__name__)
    daemon.start()

  while(True):
    time.sleep(300)
    debug("jobs queue size", JobQueueDaemon.queue.size())
    debug("running_jobs queue size", MediaJobState.queue().size())
    # debug("running_jobs queue", MediaJobState.queue().data())



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import os
import time
import datetime

import numpy as np

from sciveo.tools.logger import *
from sciveo.tools.daemon import __upload_content__ as UPLOAD
from sciveo.tools.formating import format_elapsed_time
from sciveo.tools.hardware import HardwareInfo
from sciveo.common.configuration import Configuration
from sciveo.common.model import ModelInfo
from sciveo.content.dataset import Dataset
from sciveo.api.base import APIRemoteClient


class Experiment:
  def __init__(self, project_name, project_guid, config):
    self.project_name = project_name
    self.project_guid = project_guid
    self.config = config
    self.guid = None

    self.default_score = -1e100 # TODO: It will be best to have None for default score and loss
    self.default_loss = 1e100

    self.start_at = time.time()

    self.data = {
      "experiment": {
        "log": [],
        "config": self.config(),
        "eval": {
          "score": self.default_score,
          "loss": self.default_loss,
        },
        "compute": HardwareInfo()(),
        "plots": {},
        "start_at": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
      }
    }

    # TODO: Consider better experiment naming
    # perhaps include some eval
    self.name = self.config.name
    debug("init", self.name)

  def on_guid(self, guid):
    self.guid = guid

  def close(self):
    self.end_at = time.time()
    self.elapsed = self.end_at - self.start_at

    self.data["experiment"]["end_at"] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    self.data["experiment"]["elapsed"] = format_elapsed_time(self.elapsed)
    self.data["experiment"]["dataset"] = Dataset.get().info
    debug("close", self.name)

  def append(self, d):
    self.data["experiment"]["log"].append(d)

  # EVAL API
  def eval(self, key, value, **kwargs):
    self.data["experiment"]["eval"][key] = value
    for k, v in kwargs.items():
      self.data["experiment"]["eval"][k] = v
  # Score value (no matter of its true meaning) which is used for experiments sorting.
  def score(self, value):
    score_value =  1 / (1 + np.exp(-value)) # [0.0;1.0]
    self.eval("score", score_value)
    self.eval("loss", (1.0 / score_value) - 1.0)
  # Loss value (inverse of the score in general meaning) loss=[0.0;1.0]
  def loss(self, value):
    self.eval("loss", value)
    self.score(1.0 / (1.0 + value))

  # Model API
  def model(self, m):
    self.data["experiment"]["model"] = ModelInfo(m).any()
  def model_keras(self, m):
    self.data["experiment"]["model"] = ModelInfo(m).keras()
  def model_torch(self, m):
    self.data["experiment"]["model"] = ModelInfo(m).torch()

  # LOG API
  def log(self, data, val=None, *args, **kwargs):
    if isinstance(data, dict):
      self.log_dict(data)
    else:
      self.log_any(data, val, *args, **kwargs)
  def log_dict(self, data):
    self.append(data)
  def log_any(self, data, val=None, *args, **kwargs):
    if val is None:
      self.append(data)
    else:
      self.append({data: val})
    for a in args:
      self.append(a)
    for k, v in kwargs.items():
      self.append({k: v})

  # Charts and Plots API
  def plot(self, name, data, render=None):
    # if isinstance(data, pd.DataFrame):
    #   self.data["experiment"]["plots"][name] = {"df": data.to_dict(orient='split')}
    if isinstance(data, dict):
      self.data["experiment"]["plots"][name] = {"dict": self.check_plot_x(data)}
    elif isinstance(data, list):
      # TODO: Append X as sequence if just Y is provided
      self.data["experiment"]["plots"][name] = {"list": {data[0][0]: data[0][1], data[1][0]: data[1][1]}}

    if render:
      self.data["experiment"]["plots"][name]["render"] = render

  def check_plot_x(self, data):
    # if "X" in data:
    #   if isinstance(data["X"][0], pd.Timestamp):
    #     data["X"] = [ts.strftime('%Y-%m-%d %H:%M:%S') for ts in data["X"]]
    return data

  # Datasets API
  def dataset(self, d):
    return Dataset.get(d)


class RemoteExperiment(Experiment):
  def __init__(self, project_name, project_guid, config):
    super().__init__(project_name, project_guid, config)

    self.api = APIRemoteClient()
    self.uploader = None

    remote_data = {
      "name": self.name,
      "data": self.data
    }
    if self.project_guid is not None:
      remote_data["project_guid"] = self.project_guid
    else:
      remote_data["project_name"] = self.project_name
    result = self.api.POST_SCI("experiment", remote_data)
    debug("init", self.name, "api", result)
    if result and "error" not in result:
      if result["name"] == self.name:
        self.on_guid(result["guid"])
      else:
        error("Project name mismatch", result, self.name)
    else:
      error("init", self.name, "api", result)

  def close(self):
    super().close()
    remote_data = {
      "guid": self.guid,
      "project_guid": self.project_guid,
      "data": self.data
    }
    result = self.api.POST_SCI("experiment", remote_data)
    debug("close", self.name, "api", result)

  def on_guid(self, guid):
    super().on_guid(guid)

  # UPLOAD API
  def upload(self, content_type, local_path):
    if self.guid is not None:
      UPLOAD(content_type, local_path, self.guid)
    else:
      error("upload", local_path)
  def upload_image(self, local_path):
    self.upload("image", local_path)
  def upload_file(self, local_path):
    self.upload("file", local_path)
  def upload_dataset(self, local_path):
    self.upload("dataset", local_path)



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import cv2
import datetime

from sciveo.tools.logger import *
from sciveo.media.pipelines.base import ApiContent
from sciveo.media.pipelines.processors.base import BaseProcessor


class Album(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.children = []
    self.api = ApiContent()
    self.default.update({
      "wsize": 8,
      "height": 32,
      "JPEG_QUALITY": 100,
      "channel": False
    })

    self.wsize = self["wsize"]
    self.height = self["height"]
    self.jpeg_quality = self["JPEG_QUALITY"]

  def next_content(self, media, tag, local_path):
    key = f"{tag}/{media['key']}"
    guid = f"{tag}-{media['guid']}"
    media.setdefault("next", [])
    media_next = {
      "guid": guid,
      "content_type": media['content_type'],
      "w": self.height, "h": self.height,
      "local_path": local_path,
      "key": key,
      "name": f"album {media['content_type']} {tag}",
      "bucket": media["bucket"],
      "owner": media["owner"],
      "parent": media['guid'],
      "processor": self.name(),
      "layout": {"name": self.name(), "height": self.height, **self["layout"]}
    }
    media["next"].append(media_next)
    self.children.append(media_next)

  def process(self, media):
    try:
      local_path = media["local_path"]
      frame = cv2.imread(local_path)

      tag = f"resized-{self.height}"
      resized_local_path = self.add_suffix_to_filename(local_path, tag)

      frame_resized = cv2.resize(frame, (self.height, self.height), interpolation=cv2.INTER_AREA)
      cv2.imwrite(resized_local_path, frame_resized, [cv2.IMWRITE_JPEG_QUALITY, self.jpeg_quality])

      self.next_content(media, tag, resized_local_path)
    except Exception as e:
      exception(e, self.wsize, self.height)
    return media

  def append(self, job, parent=None):
    if parent is None:
      parent = job["content_id"]

    album = {
      "guid": f"{self.name()}-{parent}",
      "parent": parent,
      "owner": job["owner"],
      "content_type": "album",
      "name": f"Album {datetime.datetime.now().strftime('%Y-%m-%d')}",
      "is_public": job["is_public"],
      "layout": {"name": self.name(), "wsize": self.wsize, "height": self.height, **self["layout"]},
      "wsize": self.wsize,
      "height": self.height,
      "children": self.children
    }
    response = self.api.create(album)
    return response

  def new_channel(self, job):
    channel = {
      "parent": job["content_id"],
      "owner": job["owner"],
      "is_public": job["is_public"],
      "content_type": "channel",
      "name": f"Album {datetime.datetime.now().strftime('%Y-%m-%d')}",
    }
    response = self.api.create(channel)

    if "error" in response:
      error("new channel failed", channel, response)
    else:
      response = self.append(job, parent=response["guid"])

    return response

  def run(self, job, input):
    next_media = super().run(job, input)

    if self["channel"]:
      self.new_channel(job)
    else:
      self.append(job)

    return next_media

  def is_resizer(self):
    return True

  def content_type(self):
    return "image"

  def name(self):
    return "image-album"

  def is_append_processor(self):
    return False



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

from sciveo.tools.logger import *
from sciveo.tools.configuration import GlobalConfiguration


class BaseML:
  def __init__(self, model_name="base", cache_dir=None, device=None) -> None:
    self.model_name = model_name
    self.config = GlobalConfiguration.get()

    if cache_dir is None:
      self.cache_dir = self.config['MODELS_BASE_PATH']
    else:
      self.cache_dir = cache_dir

    if device is None:
      self.device = os.environ.get("MEDIA_PROCESSING_BACKEND", "cpu")
    else:
      self.device = device

    self.init_models()

  def init_models(self):
    warning("init_models not implemented")

  def init(self):
    warning("init not implemented")

  def post_init(self):
    warning("post_init not implemented")

  def predict_one(self, x):
    warning("predict_one not implemented")

  def predict(self, X):
    predictions = []
    for x in X:
      predictions.append(self.predict_one(x))
    return predictions

  def train(self, X, Y_true=None):
    warning("train not implemented")

  def describe(self):
    return f"{type(self).__name__}: {self.model_name}"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import json

from sciveo.tools.logger import *


list_hyperparameters = list(set([
  # KERAS params
    'num_layers',
    'neurons_per_layer',
    'activation_function',
    'learning_rate',
    'loss_function',
    'optimizer',
    'batch_size',
    'num_epochs',
    'dropout_rate',
    'regularization_type',
    'kernel_initializer',
    'bias_initializer',
    'number_of_parameters',
    'model_type',
    'input_shape',
    'output_activation',

  # SciKit params
    'random_state',
    'n_jobs',
    'max_depth',
    'min_samples_split',
    'min_samples_leaf',
    'max_features',
    'C',
    'kernel',
    'gamma',
    'n_estimators',
    'max_samples',
    'n_neighbors',
    'weights',
    'algorithm',
    'learning_rate',
    'alpha',
    'lambda',
    'cv',
    'n_estimators',
    'n_clusters',
    'init',
    'n_components',
    'hidden_layer_sizes',
    'activation',
    'solver',
    'learning_rate_init',
]))

class Configuration:
  def __init__(self, configuration={}, predefined_attr=False):
    self.configuration = configuration
    self.name = ""

    if predefined_attr:
      for c in list_hyperparameters:
        setattr(self, c, None)
    for k, v in configuration.items():
      setattr(self, k, v)
    debug("init", configuration)

  def set(self, key, value):
    self.configuration[key] = value

  def get(self, key=None):
    if key:
      return self.configuration[key]
    else:
      return self.configuration

  def __call__(self, key=None):
    return self.get(key)

  def __getitem__(self, idx):
    return self.configuration[idx]

  def __str__(self):
    return json.dumps(self.configuration)

  def set_name(self, tag):
    self.name = tag
    for k, v in self.configuration.items():
      self.name += f" {k}={v}"
    return self.name



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import io
import base64
import numpy as np
from PIL import Image

from sciveo.tools.logger import *
from sciveo.tools.configuration import GlobalConfiguration
from sciveo.ml.base import BaseML


class BaseImageML(BaseML):
  def __init__(self, model_name=1, cache_dir=None, device=None) -> None:
    super().__init__(model_name, cache_dir, device)

  def load_image(self, x):
    if isinstance(x, Image.Image) or isinstance(x, np.ndarray):
      image = x
    elif isinstance(x, dict):
      image = Image.open(x["local_path"])
    elif isinstance(x, str):
      if x.startswith("data:image"): # Base64 encoded?
        image = Image.open(io.BytesIO(base64.b64decode(image)))
      else:
        image = Image.open(x)
    else:
      warning("unknown image format")
      image = x
    return image

  def load(self, X):
    images = []
    for x in X:
      images.append(self.load_image(x))
    return images

  def predict_one(self, x):
    return self.predict([x])[0]



#
# Stanislav Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact s.georgiev@softel.bg.
#
# 2025
#

import os
import shutil
import subprocess as sp
import time
import re
import json
import cv2
import datetime

from sciveo.tools.logger import *
from sciveo.tools.daemon import DaemonBase
from sciveo.tools.queue import TouchedFilePathQueue
from sciveo.tools.simple_counter import RunCounter


class VideoCameraCaptureDaemon(DaemonBase):
  def __init__(self, cam_id, url, dst_path, max_video_len=60, transport="tcp", max_timeout=30):
    super().__init__()
    self.cam_id = cam_id
    self.url = url
    self.dst_path = dst_path
    self.max_video_len = max_video_len
    self.max_timeout = max_timeout
    self.transport = transport
    self.cmd = [
      "ffmpeg",
      "-rtsp_transport", self.transport,
      "-i", self.url,
      "-c", "copy",
      "-acodec", "aac",
      "-f", "segment",
      "-segment_time", f"{self.max_video_len}",
      "-reset_timestamps", "1",
      "-strftime", "1",
      f"{self.dst_path}/{self.cam_id}___%Y-%m-%d___%H-%M-%S.mp4"
    ]

  def clear(self):
    os.system(f"pgrep -f \"{self.url}\" |xargs kill -9")

    files = [f for f in os.listdir(self.dst_path) if os.path.isfile(os.path.join(self.dst_path, f))]
    for file_name in files:
      if file_name.startswith(f"{self.cam_id}___"):
        file_path = os.path.join(self.dst_path, file_name)
        info("RM", file_path)
        os.remove(file_path)

  def loop(self):
    info("start", self.cam_id)

    while True:
      self.clear()

      try:
        last_known_file = None
        last_mod_time = None
        last_progress_time = time.time()

        p = sp.Popen(self.cmd, stdout=sp.DEVNULL, stderr=sp.DEVNULL)

        while True:
          time.sleep(5)

          if p.poll() is not None:
            break

          current_files = [
            f for f in os.listdir(self.dst_path)
            if f.startswith(f"{self.cam_id}___") and f.endswith(".mp4")
          ]
          if current_files:
            current_files.sort(key=lambda x: os.path.getmtime(os.path.join(self.dst_path, x)), reverse=True)
            newest_file = os.path.join(self.dst_path, current_files[0])

            mod_time = os.path.getmtime(newest_file)
            if newest_file == last_known_file:
              if mod_time > last_mod_time:
                last_progress_time = time.time()
              elif time.time() - last_progress_time > self.max_timeout:
                warning(self.cam_id, f"No file update for over {self.max_timeout} seconds, killing ffmpeg...")
                p.kill()
                break
            else:
              last_known_file = newest_file
              last_mod_time = mod_time
              last_progress_time = time.time()

      except Exception as e:
        exception(e)

      self.clear()
      warning(self.cam_id, "streaming interrupted, wait to retry...")
      time.sleep(5)



class VideoRecorder:
  def __init__(self, path_configuration):
    with open(path_configuration, 'r') as fp:
      self.configuration = json.load(fp)
    self.cams = []

    for cam_id, cam_config in self.configuration["cam"].items():
      cam = VideoCameraCaptureDaemon(
        cam_id, cam_config["url"],
        self.configuration["path"]["tmp"],
        self.configuration.get("max_video_len", 60),
        self.configuration.get("transport", "tcp")
      )
      self.cams.append(cam)

    self.queue = TouchedFilePathQueue(self.configuration["path"]["tmp"], period=5, touched_timeout=5)
    self.cleaner_timer = RunCounter(1000, self.clean_old_videos)

  def start(self):
    for cam in self.cams:
      cam.start()

    time.sleep(10)

    while(True):
      try:
        file_name, file_path = self.queue.pop()
        debug("pop", file_name, file_path)
        self.process_file(file_name, file_path)
        self.cleaner_timer.run()
      except Exception as e:
        exception(e)
        time.sleep(1)

  def process_file(self, file_name, file_path):
    split = file_name.split("___")
    if len(split) == 3:
      cam_id = split[0]
      video_date = split[1]
      video_file_name = split[2]
    else:
      warning("wrong file format, removing", file_name, file_path)
      os.remove(file_path)

    match = re.match(r"(\d{2})\-(\d{2})\-(\d{2})\.mp4", video_file_name)
    if not match:
      warning("Invalid filename format")
      video_file_name_split = video_file_name.split(".")
      video_file_name = f"{video_file_name_split[0]}-{video_file_name_split[0]}.{video_file_name_split[1]}"
    else:
      hh, mm, ss = map(int, match.groups())
      start_time = datetime.datetime(2000, 1, 1, hh, mm, ss)
      end_time = start_time + datetime.timedelta(seconds=self.configuration["max_video_len"])
      video_file_name = f"{start_time.strftime('%H.%M.%S')}-{end_time.strftime('%H.%M.%S')}.mp4"

    video_base_path = os.path.join(self.configuration["path"]["video"], cam_id, video_date)
    video_file_path = os.path.join(video_base_path, video_file_name)

    debug("MV", file_path, "=>", video_file_path)
    os.makedirs(video_base_path, exist_ok=True)
    shutil.move(file_path, video_file_path)

  def clean_old_videos(self):
    try:
      days = self.configuration.get("video_retention_period", 7)
      cmd = "find {} -mtime +{} -type f -delete".format(self.configuration["path"]["video"], days)
      debug("cmd", cmd)
      os.system(cmd)
    except Exception as e:
      excprint([self], e, cmd)


if __name__ == '__main__':
  VideoRecorder("./cams.json").start()


#!/usr/bin/env python
#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import time
import json
import argparse

from sciveo.tools.logger import *
from sciveo.tools.configuration import GlobalConfiguration


def main():
  config = GlobalConfiguration.get()

  parser = argparse.ArgumentParser(description='sciveo CLI')
  parser.add_argument(
    'command',
    choices=[
      'init', 'monitor', 'scan', 'nvr', 'predictors-server',
      'media-server', 'media-run',
      'watchdog'
    ],
    help='Command to execute')

  parser.add_argument('--period', type=int, default=120, help='Period in seconds')
  parser.add_argument('--block', type=bool, default=True, help='Block flag')
  parser.add_argument('--auth', type=str, default=config['secret_access_key'], help='Auth secret access key')
  parser.add_argument('--timeout', type=float, default=1.0, help='Timeout')
  parser.add_argument('--net', type=str, default=None, help='Network like 192.168.10.0/24')
  parser.add_argument('--host', type=str, default=None, help='Host ip or name')
  parser.add_argument('--port', type=int, default=22, help='Host port number, used for network ops')
  parser.add_argument('--ports', type=str, default="[]", help='Host ports list')
  parser.add_argument('--localhost', type=bool, default=False, help='Add localhost to list of hosts')
  parser.add_argument('--input-path', type=str, help='Input Path')
  parser.add_argument('--output-path', type=str, default=None, help='Output Path')
  parser.add_argument('--width', type=str, default=None, help='width')
  parser.add_argument('--height', type=str, default=None, help='height')
  parser.add_argument('--rate', type=int, help='Rate number')
  parser.add_argument('--processor', type=str, help='Processor name')
  parser.add_argument('--src', type=str, default=None, help='Source')
  parser.add_argument('--dst', type=str, default=None, help='Destination')
  parser.add_argument('--value', type=float, help='Value')
  parser.add_argument('--threshold', type=float, help='Threshold')
  parser.add_argument('--execute', type=str, default=None, help='Execute command')

  args = parser.parse_args()

  if args.command == 'monitor':
    from sciveo.monitoring.start import MonitorStart
    MonitorStart(period=args.period, block=args.block, output_path=args.output_path)()
  elif args.command == 'scan':
    from sciveo.network.tools import NetworkTools
    host=args.host
    if host is None:
      NetworkTools(timeout=args.timeout, localhost=args.localhost).scan_port(port=args.port, network=args.net)
    else:
      NetworkTools(timeout=args.timeout, ports=json.loads(args.ports)).scan_host(host)
  elif args.command == 'init':
    home = os.path.expanduser('~')
    base_path = os.path.join(home, '.sciveo')
    if not os.path.exists(base_path):
      os.makedirs(base_path)
      default_lines = [
        "secret_access_key=<your secret access key>",
        "api_base_url=https://sciveo.com",
        "sci_log_level=DEBUG"
      ]
      with open(os.path.join(base_path, "default"), 'w') as fp:
        for line in default_lines:
          fp.write(line + '\n')
    else:
      info(f"init, [{base_path}] already there")
  elif args.command == 'nvr':
    from sciveo.media.tools.nvr import VideoRecorder
    VideoRecorder(args.input_path).start()
  elif args.command == 'media-server':
    from sciveo.media.pipelines.server import __START_SCIVEO_MEDIA_SERVER__
    __START_SCIVEO_MEDIA_SERVER__()
  elif args.command == 'media-run':
    if args.processor == "audio-plot":
      from sciveo.media.pipelines.processors.audio.audio_extractor_process import plot_audio
      plot_audio(args.width, args.height, args.rate, args.input_path, args.output_path)
  elif args.command == 'watchdog':
    from sciveo.monitoring.watchdog.memory import MemoryWatchDogDaemon
    daemons = []
    if args.execute is not None:
      daemons.append(MemoryWatchDogDaemon(threshold_percent=args.threshold, period=args.period, command=args.execute))
    for daemon in daemons:
      daemon.start()
    while(True):
      time.sleep(3600)
  elif args.command == 'predictors-server':
    GlobalConfiguration.set("API_PREDICTORS", None)
    from sciveo.api.server import WebServerDaemon
    daemons = [
      WebServerDaemon(port=args.port)
    ]
    for daemon in daemons:
      debug("starting", type(daemon).__name__)
      daemon.start()
    while(True):
      time.sleep(3600)
  else:
    warning(args.command, "not implemented")

if __name__ == '__main__':
    main()


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import os
import time
import datetime

import numpy as np

from sciveo.tools.logger import *


class ModelInfo:
  def __init__(self, model):
    self.model = model
    self.info = {}

  def any(self):
    try:
      from tensorflow import keras
      import torch
      from sklearn.base import BaseEstimator

      if isinstance(self.model, keras.models.Model):
        return self.keras()
      elif isinstance(self.model, torch.nn.Module):
        return self.torch()
      elif isinstance(self.model, BaseEstimator):
        return self.scikit_info()
      else:
        return self.info
    except Exception as e:
      error(e)
    return self.info

  def keras(self):
    try:
      model_summary_list = []
      self.model.summary(print_fn=lambda x: model_summary_list.append(x))
      self.info = {
        "summary": self.html_from_list(model_summary_list[3:-1]),
      }
    except Exception as e:
      error(e)
    return self.info

  def scikit(self):
    try:
      for key, value in self.model.get_params().items():
        self.info[key] = value
    except Exception as e:
      error(e)
    return self.info

  def torch(self):
    return self.info

  def html_from_list(self, list_lines):
    html = ""
    for line in list_lines:
      html += f"<div>{line}</div>"
    return html

  def __call__(self):
    return self.info



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

def format_elapsed_time(seconds):
  time_units = [('d', 86400), ('h', 3600), ('m', 60), ('s', 1)]
  components = []
  for unit, factor in time_units:
    count = seconds // factor
    if count > 0:
      components.append(f"{int(count)}{unit}")
    seconds %= factor
  formatted_time = ' '.join(components)
  return formatted_time


def format_memory_size(size_in_bytes):
  if size_in_bytes < 1024:
    return f"{size_in_bytes} B"
  elif size_in_bytes < 1024 ** 2:
    return f"{size_in_bytes / 1024:.2f} KB"
  elif size_in_bytes < 1024 ** 3:
    return f"{size_in_bytes / (1024 ** 2):.2f} MB"
  elif size_in_bytes < 1024 ** 4:
    return f"{size_in_bytes / (1024 ** 3):.2f} GB"
  elif size_in_bytes < 1024 ** 5:
    return f"{size_in_bytes / (1024 ** 4):.2f} TB"
  else:
    return f"{size_in_bytes / (1024 ** 5):.2f} PB"


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import time

from sciveo.tools.logger import *
from sciveo.tools.daemon import DaemonBase
from sciveo.tools.os import *
from sciveo.tools.simple_counter import RunCounter
from sciveo.tools.configuration import ConfigurationArguments
from sciveo.tools.aws.priority_queue import SQSPriorityQueue
from sciveo.media.pipelines.pipeline import *
from sciveo.media.pipelines.queues import MediaJobState


class MaintenanceDaemon(DaemonBase):
  def __init__(self, **kwargs):
    self.arguments = ConfigurationArguments({
      "period": 3600,
      "run_period_files": 24,
      "retention_period_files": 30,
    }, **kwargs)

    super().__init__(period=self.arguments["period"])

    self.base_tmp_path = os.environ["WEB_MEDIA_PIPELINES_TMP_PATH"]

    self.timers = [
      RunCounter(self.arguments['run_period_files'], lambda: run_system_cmd(f"find {self.base_tmp_path} -mtime +{self.arguments['retention_period_files']} -type f -delete")),
      # RunCounter(3, self.print_me),
    ]

  def loop(self):
    for timer in self.timers:
      timer.run()


class MediaJobQueueDaemon(DaemonBase):
  def __init__(self):
    super().__init__()
    self.queue_level = int(os.environ.get("WEB_MEDIA_QUEUE_LEVEL", 1))
    self.queue_id = int(os.environ.get("WEB_MEDIA_QUEUE_ID", 1))
    self.queue = SQSPriorityQueue(self.queue_level, self.queue_id)

  def loop(self):
    job = self.queue.pull_wait()

    debug("new process job", job)
    job["input"] = job["content"]
    MediaJobState.queue().start(job["id"])
    MediaJobState.queue().inc_progress(job["id"], 10)

    pipeline = MediaPipeline(job)
    job = pipeline.run()

    MediaJobState.queue().finish(job["id"], job)


def __START_SCIVEO_MEDIA_SERVER__():
  daemons = [
    MaintenanceDaemon(period=3600, run_period_files=24, retention_period_files=30),
    MediaJobQueueDaemon(),
  ]

  for daemon in daemons:
    daemon.start()

  while(True):
    time.sleep(12 * 3600)

if __name__ == "__main__":
  __START_SCIVEO_MEDIA_SERVER__()



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import numpy as np
import cv2

from sciveo.tools.common import *
from sciveo.media.pipelines.processors.base import *


class ImageWatermark(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.default.update({"JPEG_QUALITY": 100, "watermark": "SMIVEO", "height": 480, "thickness": 10, "font_scale": 3.0, "angle": 30})

  def apply(self, frame, image_local_path):
    color = (0, 0, 255)
    thickness = self["thickness"]
    font_scale = self["font_scale"]
    angle = self["angle"]
    font = cv2.FONT_HERSHEY_TRIPLEX
    alpha = 0.7
    noise_scale = 1.0

    (text_width, text_height) = cv2.getTextSize(self["watermark"], font, fontScale=font_scale, thickness=thickness)[0]

    w = frame.shape[1]
    h = frame.shape[0]

    # Determine the position of the watermark text
    (x, y) = (int((w - text_width)/2), int(h * 0.5))

    # Create an empty overlay image with the same size as the original image
    overlay = np.zeros_like(frame)

    # Add the watermark text to the overlay image with rotation
    cv2.putText(overlay, self["watermark"], (x, y), font, fontScale=font_scale, color=color, thickness=thickness, lineType=cv2.LINE_AA)
    (overlay_height, overlay_width) = overlay.shape[:2]
    center = (overlay_width // 2, overlay_height // 2)
    M = cv2.getRotationMatrix2D(center, angle, 1)
    overlay = cv2.warpAffine(overlay, M, (overlay_width, overlay_height))

    # Create a mask of the watermark text
    gray_overlay = cv2.cvtColor(overlay, cv2.COLOR_BGR2GRAY)
    ret, mask = cv2.threshold(gray_overlay, 0, 255, cv2.THRESH_BINARY)
    mask = cv2.merge([mask, mask, mask])

    # Add Gaussian noise to the watermark text
    noise = np.zeros_like(overlay)
    cv2.randn(noise, 0, 255 * noise_scale)
    # Apply the mask to the noise image
    noise = cv2.bitwise_and(noise, mask)
    # Blend the noise image with the overlay image
    cv2.addWeighted(overlay, 0.2, noise, 0.8, 0, overlay)

    # Add the overlay image to the original image with alpha blending
    cv2.addWeighted(overlay, alpha, frame, 1, 0, frame)
    cv2.imwrite(image_local_path, frame, [cv2.IMWRITE_JPEG_QUALITY, self["JPEG_QUALITY"]])

  def process(self, media):
    try:
      self.media = media
      self.local_path = media["local_path"]
      frame = cv2.imread(self.local_path)

      tag = "WTRMRK"
      image_local_path = self.add_suffix_to_filename(self.local_path, tag)

      h, w = get_frame_resolution(frame, self["height"])
      frame_resized = cv2.resize(frame, (w, h), interpolation=cv2.INTER_AREA)

      self.apply(frame_resized, image_local_path)
      self.next_content(self.media, tag, image_local_path, w=frame_resized.shape[1], h=frame_resized.shape[0])
    except Exception as e:
      exception(e, self.media)
    return self.media

  def content_type(self):
    return "image"

  def name(self):
    return "image-watermark"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import cv2
import math
import numpy as np

from sciveo.tools.logger import *
from sciveo.tools.common import *
from sciveo.media.pipelines.processors.base import *
from sciveo.media.pipelines.queues import MediaJobState


class AlbumInImage(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.default.update({
      "max_col": 10,
      "width": 640,
      "height": 640,
      "bucket": "smiveo-video"
    })

  def create_album_image(self, images, width, height, max_col=10):
    num_images = len(images)
    num_cols = min(max_col, math.ceil(np.sqrt(num_images)))
    num_rows = (num_images + num_cols - 1) // num_cols
    w = width // num_cols
    h = height // num_rows
    debug("create_album_image", "num_images", num_images, "num_cols", num_cols, "num_rows", num_rows, "WH", (w, h))
    resized_images = [cv2.resize(image, (w, h), interpolation=cv2.INTER_AREA) for image in images]
    canvas = np.zeros((height, width, 3), dtype=np.uint8)
    for i, image in enumerate(resized_images):
      x = (i % num_cols) * w
      y = (i // num_cols) * h
      canvas[y:y + image.shape[0], x:x + image.shape[1], :] = image
    return canvas

  def run(self, job, input):
    if len(input) == 0:
      return []

    progress_per_media = self.max_progress / len(input)
    debug("run", job["id"], progress_per_media)

    images = []
    videos = []

    # TODO: Handle different resolutions and aspect ratios
    for i, media in enumerate(input):
      if media["content_type"] == "image":
        images.append(media)
      if media["content_type"] == "video":
        videos.append(media)

    max_col = self["max_col"]
    width = self["width"]
    height = self["height"]

    list_frames = []

    for media in images:
      try:
        frame = cv2.imread(media["local_path"])
        list_frames.append(frame)
      except Exception as e:
        exception(e)

    for media in videos:
      try:
        cap = cv2.VideoCapture(media["local_path"])
        frames_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        frame_idx = frames_count // 2
        cap.set(cv2.CAP_PROP_POS_FRAMES, frame_idx)
        ret, frame = cap.read()
        if ret:
          list_frames.append(frame)
        cap.release()
      except Exception as e:
        exception(e)
      MediaJobState.queue().inc_progress(job["id"], progress_per_media)

    parent_guid = job['content_id']
    guid = f"{self.name()}-{parent_guid}"
    key = f"{guid}.jpg"
    bucket = self["bucket"]

    album_image = self.create_album_image(list_frames, width, height, max_col=max_col)
    album_image_local_path = os.path.join(self.base_tmp_path, key)
    cv2.imwrite(album_image_local_path, album_image, [cv2.IMWRITE_JPEG_QUALITY, 100])

    media = {
      "guid": guid,
      "parent": parent_guid,
      "content_type": "image",
      "owner": job["owner"],
      "local_path": album_image_local_path,
      "w": width, "h": height,
      "height": height,
      "key": key,
      "bucket": bucket,
      "processor": self.name(),
      "layout": {"name": self.name(), "height": height, **self["layout"]}
    }

    if self["output"]:
      job["output"].append(media)
    if self["append-content"]:
      job["append-content"].append(media)
    return [media]

  def content_type(self):
    return "image"

  def name(self):
    return "album-in-image"

  def is_append_processor(self):
    return False


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import re
import datetime
import os
import os.path
import json
import cv2
import skvideo.io

from sciveo.tools.logger import *


def get_resize_resoulution(FH, FW, max_dim):
  if FH <= max_dim:
    debug("get_resize_resoulution", [FH, FW], max_dim)
    return FH, FW
  h = max_dim
  w = int(FW * h / FH)
  if w % 2 == 1:
    w += 1
  return h, w

def get_frame_resolution_max_dim(frame, max_dim):
  FH = frame.shape[0]
  FW = frame.shape[1]

  if FH > FW:
    return get_resize_resoulution(FH, FW, max_dim)
  else:
    w, h = get_resize_resoulution(FW, FH, max_dim)
    return h, w

def get_frame_resolution(frame, h):
  FH = frame.shape[0]
  FW = frame.shape[1]
  return get_resize_resoulution(FH, FW, h)


class VideoWriterBase:
  def __init__(self, local_video_path, w=640):
    self.local_video_path = local_video_path
    self.w = w
    self.draw_frame = True

  def use_frame_draw(self):
    return self.draw_frame

  def write(self, frame, detections=None):
    pass

  def close(self):
    pass

class VideoWriterDummy(VideoWriterBase):
  def __init__(self, local_video_path, w=640):
    super().__init__(local_video_path, w)
    debug("init", self.local_video_path)

  def write(self, frame, detections=None):
    pass

  def close(self):
    debug("close", self.local_video_path)
    pass

class VideoWriterFFMPEG(VideoWriterBase):
  def __init__(self, local_video_path, w=640, crf='35', aux_params={}):
    super().__init__(local_video_path, w)

    output_params = {'-vcodec':'libx264', '-crf':crf, '-preset':'medium', '-vf':'format=yuv420p', '-profile:v':'baseline', '-level':'3.0'}
    output_params = {**output_params, **aux_params}
    info(self.local_video_path, output_params)
    self.out = skvideo.io.FFmpegWriter(self.local_video_path, outputdict=output_params)

  def write(self, frame, detections=None):
    self.out.writeFrame(frame[:,:,::-1])

  def close(self):
    self.out.close()
    debug("closed", self.local_video_path)

  @staticmethod
  def new(local_video_path, w=640, crf='35'):
    if int(os.environ.setdefault('PRODUCTION_MODE', "1")):
      result = VideoWriterFFMPEG(local_video_path, w, crf)
    else:
      result = VideoWriterDummy(local_video_path, w)
    return result



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import gc
import torch
from sciveo.tools.logger import *
from sciveo.media.pipelines.processors.base import *


class TPUBaseProcessor(BaseProcessor):
  def post_process(self):
    try:
      del self.model
    except:
      pass
    try:
      del self.pipe
    except:
      pass
    self.clear()

  def clear(self):
    debug("clear, cuda available", torch.cuda.is_available())
    if torch.cuda.is_available():
      gc.collect()
      torch.cuda.empty_cache()
      torch.cuda.ipc_collect()
      debug("torch.cuda cleared")



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import cv2

from sciveo.tools.common import *
from sciveo.media.pipelines.processors.base import *


class VideoDownsample(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.videos = []
    self.downsampling_rates = []
    self.default.update({"CRF": 20, "downsampling-rates": [10, 20]})

  def init_videos(self, frame, media):
    for downsampling_rate in self["downsampling-rates"]:
      h, w = frame.shape[0], frame.shape[1]
      dsr = f"DSR{downsampling_rate}"
      video_local_path = self.add_suffix_to_filename(self.local_path, dsr)
      self.downsampling_rates.append(downsampling_rate)
      self.videos.append(VideoWriterFFMPEG.new(video_local_path, crf=str(self["CRF"])))
      self.next_content(media, dsr, video_local_path, w=w, h=h)

  def process(self, media):
    self.local_path = media["local_path"]

    self.cap = cv2.VideoCapture(self.local_path)
    n_frame = 0
    while(self.cap.isOpened()):
      n_frame += 1
      ret, frame = self.cap.read()
      if not ret:
        break

      if n_frame == 1:
        self.init_videos(frame, media)

      for i, video in enumerate(self.videos):
        if n_frame % self.downsampling_rates[i] == 0:
          try:
            video.write(frame)
          except Exception as e:
            exception(e, i, self.downsampling_rates[i])

    self.cap.release()
    for video in self.videos:
      video.close()

    return media

  def content_type(self):
    return "video"

  def name(self):
    return "video-downsample"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import numpy as np
import cv2

from sciveo.tools.logger import *
from sciveo.tools.simple_counter import Timer
from sciveo.tools.common import *
from sciveo.ml.images.object_detection import ObjectDetectorYOLO
from sciveo.media.pipelines.processors.base import *


class ImageObjectDetectionProcessor(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.default.update({"JPEG_QUALITY": 80, "min_confidence": 0.5, "model_type": 0, "height": 720})

  def init_run(self):
    self.predictor = ObjectDetectorYOLO(model_path=[
      "yolo11x.pt", "yolo11l.pt", "yolo11m.pt", "yolo11s.pt", "yolo11n.pt",
      'softel-surveillance-yolo11X.pt', 'softel-surveillance-yolo11L.pt', 'softel-surveillance-yolo11M.pt',
      'softel-surveillance-yolo11S.pt', 'softel-surveillance-yolo11N.pt',
    ][self["model_type"]])

  def process(self, media):
    try:
      self.media = media
      self.local_path = media["local_path"]

      tag = "object-detections"
      image = self.predictor.load(self.local_path)

      image_resized = self.predictor.resize(image)
      detections = self.predictor.predict_one([image_resized], confidence_threshold=self["min_confidence"])

      image_resized = self.predictor.resize(image, h=self["height"])
      image_resized = cv2.cvtColor(image_resized, cv2.COLOR_RGB2BGR)
      self.predictor.draw(image_resized, detections[0])
      result_image_local_path = self.add_suffix_to_filename(self.local_path, tag)
      cv2.imwrite(result_image_local_path, image_resized, [cv2.IMWRITE_JPEG_QUALITY, self["JPEG_QUALITY"]])

      self.next_content(self.media, tag, result_image_local_path, w=image_resized.shape[1], h=image_resized.shape[0])
      self.next_comment(detections[0])
    except Exception as e:
      exception(e, self.media)
    return self.media

  def next_comment(self, detections):
    boxes = detections.boxes
    class_names = detections.names
    detections_json = {}
    for i, box in enumerate(boxes):
      confidence = box.conf[0].item()
      class_id = int(box.cls[0].item())
      label = class_names[class_id]
      detections_json.setdefault(label, 0)
      detections_json[label] += 1

    detections_str = ""
    for k, v in detections_json.items():
      detections_str += f"{v} {k}\n"

    self.next_content(self.media, tag="OD", content_type="comment", name=f"OD {self.predictor.model_name}", args={
      "description": detections_str,
      "content_text": f"{detections_str}\n\n{str(detections_json)}"
    })

  def content_type(self):
    return "image"

  def name(self):
    return "image-object-detection"



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

from sciveo.tools.logger import *

from sciveo.media.pipelines.processors.audio.audio import *

from sciveo.media.pipelines.processors.video.video_resample import *
from sciveo.media.pipelines.processors.video.video_frames import *
from sciveo.media.pipelines.processors.video.motion_detection import *
from sciveo.media.pipelines.processors.video.resize import *
from sciveo.media.pipelines.processors.video.video_album import *
from sciveo.media.pipelines.processors.video.generators import *

from sciveo.media.pipelines.processors.image.histogram import *
from sciveo.media.pipelines.processors.image.resize import *
from sciveo.media.pipelines.processors.image.album import *
from sciveo.media.pipelines.processors.image.album_in_image import *
from sciveo.media.pipelines.processors.image.filters import *
from sciveo.media.pipelines.processors.image.generators import *
from sciveo.media.pipelines.processors.image.embeddings import *
from sciveo.media.pipelines.processors.image.depth_esimation import *
from sciveo.media.pipelines.processors.image.mask import *
from sciveo.media.pipelines.processors.image.segmentation import *
from sciveo.media.pipelines.processors.image.watermark import *
from sciveo.media.pipelines.processors.image.object_detection import ImageObjectDetectionProcessor

from sciveo.media.pipelines.processors.file.archive import *

from sciveo.media.pipelines.processors.sci.time_series.predictor import *
from sciveo.media.pipelines.processors.sci.time_series.trainer import *
from sciveo.media.pipelines.processors.sci.dataset import *

from sciveo.media.pipelines.processors.nlp.address import *

from sciveo.media.pipelines.processors.aws import *
from sciveo.media.pipelines.processors.media_info import *
from sciveo.media.pipelines.processors.qr import *

from sciveo.media.pipelines.postprocessors.default import *


class MediaPipeline:
  def __init__(self, job) -> None:
    self.job = job
    self.configuration = job["configuration"]
    debug("configuration", self.configuration)
    self.pipeline = []

    # TODO: Should have subscription-based processors list instead
    # TODO: Replace/simplify with a list of classes only and build this map
    self.available_processors = {
      "s3-download": S3MediaDownload,
      "archive-zip": FileArchiveZIP,
      "media-info": MediaInfo,
      "QR-generator": QRGenerator,

      "audio-extract": AudioExtract,

      "video-resize": VideoResize,
      "video-downsample": VideoDownsample,
      "video-frames-extract": VideoFramesExtract,
      "video-motion-detector": VideoMotionDetector,
      "video-album": VideoAlbum,
      "video-to-text": VideoToTextProcessor,

      "image-resize": ImageResize,
      "image-histogram": ImageHistogram,
      "image-album": Album,
      "album-in-image": AlbumInImage,
      "image-filters": ImageFilters,
      "image-watermark": ImageWatermark,
      "image-diffusion": ImageDiffusionText,
      "image-diffusion-image-text": ImageDiffusionImageText,
      "image-to-text": ImageToTextProcessor,
      "image-fgbg-filter": ImageFGBGFilter,
      "image-segmentation": ImageSegmentation,
      "image-depth-estimation": ImageDepthEstimation,
      "image-embedding": ImageEmbeddingProcessor,
      "image-object-detection": ImageObjectDetectionProcessor,

      "sci-timeseries-predictor": TimeSeriesPredictorProcessor,
      "sci-timeseries-trainer": TimeSeriesTrainerProcessor,
      "project-datasets-plots": ProjectDatasetPlots,

      "nlp-address-standard": AddressStandard,
      "nlp-address-geocode": AddressGeocode,
    }

    self.available_postprocessors = {
      "s3-upload": S3UploadPostprocessor,
      "append-content": AppendContentPostprocessor,
      "update-content-data": UpdateContentDataPostprocessor,
      "resized-resolutions": ResizedResolutionsPostprocessor,
      "channel-create-simple": ChannelCreateSimplePostprocessor
    }

    self.count_processors = 0
    self.count_processors_dfs(0, self.configuration["processors"])
    debug("count_processors", self.count_processors)

  def describe(self):
    d = {"processors": {}, "postprocessors": {}}
    for processor_name, processor in self.available_processors.items():
      d["processors"][processor_name] = processor({}, 100).describe()
    for postprocessor_name, postprocessor in self.available_postprocessors.items():
      d["postprocessors"][postprocessor_name] = postprocessor(self.job, {}, 100).describe()
    return d

  def count_processors_dfs(self, level, processors):
    for processor_name, processor_definition in processors.items():
      debug(level, processor_name)
      self.count_processors += 1
      if "next" in processor_definition:
        self.count_processors_dfs(level + 1, processor_definition["next"])

  # Walk processors and filter input
  def walk_processors_dfs(self, level, input, processors, result, max_level=1):
    if level > max_level:
      return

    for processor_name, processor_definition in processors.items():
      # debug("walk", level, processor_name)

      processor = self.available_processors[processor_name](processor_definition.get("configuration", {}), 100)

      for i, media in enumerate(input):
        # Check if the media really needs this processor, add to input if should be processed by it.
        if processor.is_processor_run(self.job, media) and processor.is_processor_output():
          debug("walk add media", media['guid'], media.get('processors', ""), "because of", processor.append_name(), [processor.is_processor_run(self.job, media), processor.is_processor_output()])
          result[media['guid']] = media

      if "next" in processor_definition:
        self.walk_processors_dfs(level + 1, input, processor_definition["next"], result)

  def run_processors_dfs(self, level, input, processors):
    for processor_name, processor_definition in processors.items():
      debug(level, processor_name)

      processor = self.available_processors[processor_name](processor_definition.get("configuration", {}), 40 / self.count_processors) # Account 60% for all processors
      next_input = processor.run(self.job, input)

      processor.post_process()

      if "next" in processor_definition:
        self.run_processors_dfs(level + 1, next_input, processor_definition["next"])

  def run_postprocessors(self):
    postprocessors = self.configuration.get("postprocessors", {
      "s3-upload": {}, "append-content": {}, "update-content-data": {},"resized-resolutions": {},
    })

    for postprocessor_name, postprocessor_definition in postprocessors.items():
      if postprocessor_name in self.available_postprocessors:
        postprocessor = self.available_postprocessors[postprocessor_name](self.job, postprocessor_definition.get("configuration", {}), 60 / len(postprocessors))
        postprocessor.run()
      else:
        warning("postprocessor", postprocessor_name, "is not available")

  def run(self):
    self.job["output"] = []

    for k, v in self.available_postprocessors.items():
      self.job[k] = []

    list_media = {}
    max_walk_level = self.configuration["processors"].get("max_walk_level", 1)
    self.walk_processors_dfs(0, self.job["input"], self.configuration["processors"], list_media, max_walk_level)
    self.job["input"] = list(list_media.values())
    self.job["content"] = self.job["input"]
    debug("filtered content for processing", list(list_media.keys()))

    self.run_processors_dfs(0, self.job["input"], self.configuration["processors"])

    self.run_postprocessors()

    return self.job



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import cv2
from PIL import Image
import numpy as np

import torch
from torchvision import transforms

from sciveo.tools.os import replace_extension_to_filename
from sciveo.tools.common import *
from sciveo.media.pipelines.processors.tpu_base import *
from sciveo.media.pipelines.processors.image.filters import Filters


class ForegroundMask:
  def __init__(self):
    torch.hub.set_dir(os.path.join(os.environ['MEDIA_MODELS_BASE_PATH'], "models/"))
    self.model = torch.hub.load('pytorch/vision:v0.6.0', 'deeplabv3_resnet101', pretrained=True)
    self.model.eval().to(os.environ.get("MEDIA_PROCESSING_BACKEND", "cpu"))

  def predict(self, image):
    preprocess = transforms.Compose([
      transforms.ToTensor(),
      transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ])

    input_tensor = preprocess(Image.fromarray(image)).to(os.environ.get("MEDIA_PROCESSING_BACKEND", "cpu"))
    input_batch = input_tensor.unsqueeze(0)

    with torch.no_grad():
      output = self.model(input_batch)['out'][0]
    output_predictions = output.argmax(0)

    # create a binary (black and white) mask of the profile foreground
    mask = output_predictions.byte().cpu().numpy()
    background = np.zeros(mask.shape)
    bin_mask = np.where(mask, 255, background).astype(np.uint8)

    if image.shape[2] > 1:
      bin_mask = np.repeat(bin_mask[:, :, np.newaxis], image.shape[2], axis=2)

    del input_tensor
    del input_batch
    del output
    del output_predictions

    return bin_mask


class ImageFGBGFilter(TPUBaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.filters = Filters()
    self.mask = ForegroundMask()
    self.default.update({"JPEG_QUALITY": 100, "height": 720, "filters": {"FG": [], "BG": []}})

  def post_process(self):
    del self.mask.model
    self.clear()

  def get_filters_groups(self):
    if isinstance(self["filters"], dict):
      return [self["filters"]]
    elif isinstance(self["filters"], list):
      return self["filters"]
    else:
      return self["filters"]

  def process(self, media):
    try:
      self.media = media
      self.local_path = media["local_path"]

      image = cv2.cvtColor(cv2.imread(self.local_path), cv2.COLOR_BGR2RGB)
      h, w = get_frame_resolution(image, self["height"])
      image = cv2.resize(image, (w, h), interpolation=cv2.INTER_AREA)

      mask = self.mask.predict(image)

      filters_groups = self.get_filters_groups()
      for filters in filters_groups:
        tag = ""
        masked = {"FG": mask & image, "BG": (~mask) & image}
        for mask_name in masked.keys():
          tag += f"{mask_name}"
          for f in filters.get(mask_name, []):
            masked[mask_name] = self.filters(f[0], masked[mask_name], f[1])
            tag += f"{f[0]}{f[1]}"

        filtered = masked["FG"] | masked["BG"]

        image_local_path = self.add_suffix_to_filename(self.local_path, tag)
        image_local_path = replace_extension_to_filename(image_local_path, "png")

        cv2.imwrite(
          image_local_path,
          cv2.cvtColor(filtered, cv2.COLOR_RGB2BGR),
          [cv2.IMWRITE_JPEG_QUALITY, self["JPEG_QUALITY"]]
        )

        self.next_content(self.media, tag, image_local_path, w=w, h=h)
    except Exception as e:
      exception(e, self.media)
    return self.media

  def content_type(self):
    return "image"

  def name(self):
    return "image-fgbg-filter"

  def append_name(self):
    result = self.name()
    filters_groups = self.get_filters_groups()
    for filters in filters_groups:
      for k, list_filters in filters.items():
        result += f"-{k}"
        for f in list_filters:
          result += f"{f[0]}{f[1]}"
    return result



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import json
import base64
from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes

from sciveo.tools.logger import *


class CryptoJsonData:
  def __init__(self, base64_key=None) -> None:
    if base64_key is None:
      self.gen_key()
    else:
      self.key = base64.b64decode(base64_key)

  def get_key(self):
    return base64.b64encode(self.key).decode("utf-8")

  def gen_key(self):
    self.key = get_random_bytes(32)
    return self.get_key()

  def read_key(self, key_path):
    with open(key_path, 'rb') as fp:
      self.key = fp.read(32)

  def write_key(self, key_path):
    with open(key_path, 'wb') as fp:
      fp.write(self.key)

  def encrypt(self, plaintext, mode=AES.MODE_GCM):
    iv = get_random_bytes(12)
    cipher = AES.new(self.key, mode, nonce=iv)
    data_encrypted, auth_tag = cipher.encrypt_and_digest(plaintext.encode())
    return {
      "encrypted": base64.b64encode(data_encrypted).decode("utf-8"),
      "auth_tag": base64.b64encode(auth_tag).decode("utf-8"),
      "iv": base64.b64encode(iv).decode("utf-8")
    }

  def decrypt(self, cipher_data, mode=AES.MODE_GCM):
    cipher_text = base64.b64decode(cipher_data["encrypted"])
    auth_tag = base64.b64decode(cipher_data["auth_tag"])
    iv = base64.b64decode(cipher_data["iv"])
    cipher = AES.new(self.key,  mode, nonce=iv)
    return cipher.decrypt_and_verify(cipher_text, auth_tag).decode("utf-8")

  def encrypt_json(self, data):
    data = json.dumps(data)
    data = base64.b64encode(data.encode()).decode("utf-8")
    return self.encrypt(data)

  def decrypt_json(self, encrypted_data):
    data = self.decrypt(encrypted_data)
    data = base64.b64decode(data)
    data = json.loads(data)
    return data
